//  Copyright (c) 2002-2004, Gerben Wierda, R&A
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
//  Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//  Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//
//  Neither the name of R&A nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#import <Cocoa/Cocoa.h>
#import "GLUPTaskRunner.h"
#import "GWAuthenticator.h"
#import "II2TaskSet.h"

@class GLUPTaskRunner;

@protocol GLUPTaskOwnerProtocol
- (void)glupTaskDidFinish: (GLUPTaskRunner *)task withSuccess:(BOOL)success;
- (void)glupTaskWillStart: (GLUPTaskRunner *)task;
@end

typedef enum {
    GLUPTaskRunnerNoPid,
    GLUPTaskRunnerStdoutPid,
    GLUPTaskRunnerFifoPid
} GLUPTaskRunnerReadPidStyle;

@interface GLUPTaskRunner : NSObject <II2TaskSetMemberProtocol, II2TaskPropertiesProtocol, II2TaskStatusProviderProtocol>
{
    id <GLUPTaskOwnerProtocol>	__owner;
    NSString			*__displayName;		// owned, cache for speed
    NSAttributedString		*__attributedDisplayName;	// ditto
    BOOL			__displayNameWithPID;
    NSString			*__directoryPath;	// copy
    NSString			*__tmpDirectoryPath;	// copy
    BOOL			__keepTmpDirOnExit;
    GLUPTaskRunnerReadPidStyle	__reportsPID;
    NSMutableArray		*__arguments;		// copy
    NSTextView			*__outView;		// reference
    NSTextView			*__errView;		// reference
    NSMutableDictionary		*__environment;		// copy
    GWAuthenticator		*__authenticator;	// reference or nil
    pid_t			__pid;
    int				__result;
    NSMutableData		*__outData;		// retained/released reference
    NSMutableData		*__errData;		// retained/released reference
    BOOL			__checkOutSize;
    size_t			__outSize;
    size_t			__outSizeLimit;
    NSMutableDictionary		*__properties;		// copy
    NSFileHandle		*___authTaskInOutHandle;	// owned, auth tasks
    NSFileHandle		*___taskInHandle;	// owned, nonauth tasks
    NSFileHandle		*___taskOutHandle;	// owned, nonauth tasks
    NSFileHandle		*___taskErrHandle;	// owned, nonauth tasks
    NSFileHandle		*__taskInHandle;		// References to auth ___authTaskInOutHandle if auth task
    NSFileHandle		*__taskOutHandle;	// References to auth ___authTaskInOutHandle if auth task
    NSFileHandle		*__taskErrHandle;	// 
    GLUPTaskRunner		*__outTask; // array??	// reference
    GLUPTaskRunner		*__isOutTaskOf; 		// reference
    NSString			*__gluprunner;		// Launchpath for all processes. Gluprunner is *G*erben's process-*L*eader setting, real *U*id setting, *P*id reporting runner tool.
    NSTask			*__nonAuthTask;		// Used when no auth is given
    BOOL			__synchronous;		// Wait for result or run in background
    NSRange			__displayRange;
    II2TaskSet			*__taskSet;
    unsigned int		__waitCounter;
    BOOL			__invalidated;
    NSMutableArray		*__dependentTasks;
    NSString 			*__envFifoName;
    NSString 			*__stderrFifoName;
    NSString 			*__pidFifoName;
    NSConnection		*__kitConnection;
    BOOL			__errorPipeCreated;
    BOOL			__outputPipeCreated;
    BOOL			__inputPipeCreated;
    BOOL                        __Skip; // When set the task is not actually executed, but closed with success
    id <II2TaskStatusReporterProtocol> __reporter;
    int				__currentOutputSize;
    int				__expectedOutputSize;
    BOOL			__expectedOutputIsErrData;
    int				__reportedOutputSize;
    int				__reportIncrementTreshold;
    NSString			*__reportName;
}

+ (NSString *)fixedLaunchPath;

- (id)initWithOwner:(id <GLUPTaskOwnerProtocol>)owner reportsPID:(GLUPTaskRunnerReadPidStyle)reportsPID arguments:(NSArray *)arguments authenticator:(GWAuthenticator *)authenticator;
- (void)setEnvironmentValue:(NSString *)value forKey:(NSString *)key;
- (void)setEnvironment: (NSDictionary *)anEnvironment;
- (void)setDirectoryPath: (NSString *)aDirectoryPath;
- (void)useTmpDir: (NSString *)aDirectoryPath cleanOnFinish:(BOOL)cleanOnFinish;

- (BOOL)isAuthTask;

- (void)setErrView: (NSTextView *)aView;
- (void)setOutView: (NSTextView *)aView;
- (void)setOutData: (NSMutableData *)aData;
- (void)setErrData: (NSMutableData *)aData;
- (void)setOutSizeLimit: (size_t)limit;

- (NSString *)identifier;
- (void)run;
- (void)launch;
- (void)terminate;
- (void)abortReportFailure:(BOOL)failure;
- (int)result;
- (BOOL)isRunning;
- (pid_t)pid;

// Connecting tasks, non-auth only
// When connected, no other output (outData, outView) is possible
- (void) connectOutputTo: (GLUPTaskRunner *)task;
- (void) getInputFrom: (NSPipe *)pipe;

// Getting the PID through a fifo from gluprunner, auth only
- (void) makeFiFo: (NSString*)fifoName;
- (void) readPidFifo: (NSString*)fifoName;
- (void)doErrDataHandler: (NSData *)data;

- (BOOL)skip;
- (void)setSkip: (BOOL)skip;

- (void)setReportTarget:(id <II2TaskStatusReporterProtocol>)reporter;
- (NSString *)typeOfTask;
- (int)sizeOfTaskWhenCompleted;
- (int)currentSizeOfTask;
- (int)sizeOfTaskSinceLastReported;

@end


/*
 * $Log: GLUPTaskRunner.h,v $
 * Revision 1.3  2004/04/25 13:13:14  gctwnl
 * Forced commit of all .h and .m files of version 2.61.0
 *
 * Revision 1.2  2004/03/19 23:35:11  gctwnl
 * Version 2.55.0. Has a renewed interface to support simple operations.
 * Supports progress bar control from the i-Package.
 *
 * Revision 1.1.1.1  2004/02/01 17:17:28  gctwnl
 * Initial import.
 *
 * Revision 3.5  2004/01/14 12:56:45  gerben
 * Version 2.39.0
 *
 * Revision 3.4  2003/11/20 11:56:06  gerben
 * 2.32.0 work in progress
 *
 * Revision 3.3  2003/09/18 22:40:54  gerben
 * 2.26.0 Fragmented archives
 *
 * Revision 3.2  2002/12/10 14:24:47  gerben
 * Set sticky tag 3.2
 *
 * Revision 3.0  2002/12/10 14:04:21  gerben
 * Upping all files with a sticky tag of 1.99
 *
 * Revision 1.99  2002/12/10 13:52:49  gerben
 * Set all revision numbers to 1.99
 *
 * Revision 1.20  2002/10/22 14:14:20  gerben
 * Differ between different user actions.
 *
 * Revision 1.19  2002/10/22 12:52:21  gerben
 * Added stderr handling to authenticated tasks.
 *
 * Revision 1.18  2002/10/14 12:34:20  gerben
 * Removed dead code
 *
 * Revision 1.17  2002/10/12 10:54:39  gerben
 * A lot of changes, mostly when producing output/feedback.
 *
 * Revision 1.16  2002/10/10 00:51:25  gerben
 * Many changes. Committing a stable state.
 *
 * Revision 1.15  2002/10/09 14:22:23  gerben
 * Removed outputHandler from GLUPTaskRunner
 *
 * Revision 1.14  2002/10/09 10:53:40  gerben
 * Moved textview output to NSTexTView category.
 *
 * Revision 1.13  2002/10/08 22:25:03  gerben
 * GPG checking complete. Output of parallel tasks will not mingle.
 *
 * Revision 1.12  2002/10/07 20:28:06  gerben
 * Removed circular CVS Log entry in C Comments at the end of files
 *
 * Revision 1.11  2002/10/06 21:05:07  gerben
 * Progress Indicator for active tasks, hidden when there are no processes.
 *
 * Revision 1.10  2002/10/06 11:14:38  gerben
 * Changed displaying process information to a more friendly format.
 *
 * Revision 1.9  2002/10/02 22:38:31  gerben
 * Added CVS Log in C Comment at the end of every file
 *
 */
