//  Copyright (c) 2002-2004, Gerben Wierda, R&A
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
//  Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//  Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//
//  Neither the name of R&A nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#import "GWAuthenticator.h"

static char rcsid[] = "$Id: GWAuthenticator.m,v 1.2 2004/04/25 13:13:15 gctwnl Exp $";

@implementation GWAuthenticator

- (BOOL)resetPrivileges
{
    OSStatus err = 0;
    AuthorizationRights rights;
    AuthorizationItem items[1];

    if (__authorizationRef) {
	// Destroy these privileges, do not render them invalid for other running processes
	AuthorizationFree( __authorizationRef, 0);
	__authorizationRef = NULL;
    }

    // The authorization rights structure holds a reference to an array of AuthorizationItem structures that represent the rights for which you are requesting access.

    items[0].name = kAuthorizationRightExecute;
    items[0].value = (char *)[__launchPath fileSystemRepresentation];
    items[0].valueLength = strlen( items[0].value);
    items[0].flags = 0;

    rights.count=1;
    rights.items = items;

    err = AuthorizationCreate( &rights, kAuthorizationEmptyEnvironment, kAuthorizationFlagDefaults|kAuthorizationFlagPartialRights, &__authorizationRef);

    if (err != errAuthorizationSuccess) {
	NSLog( @"resetPrivileges: Authorization Framework failed with error: %d", err);
	__authorizationRef = NULL;
	return NO;
    }
    else {
	return YES;
    }
}

- (id)initWithLaunchPath:(NSString *)launchPath
{
    self = [super init];
    if (self) {
	// We have or own copy
	__launchPath = [[NSString alloc] initWithString:launchPath];

	// get our startup privileges
	[self resetPrivileges];

	// Have we succeeded in setting up startup conditions?
	if (!__authorizationRef) {
	    [self release];
	    return nil;
	}
    }
    return self;
}

- (void)dealloc
{
    [__launchPath release];
    if (__authorizationRef) {
	// Destroy these privileges, do not render them invalid for other running processes
	AuthorizationFree( __authorizationRef, 0);
	__authorizationRef = NULL;
    }
    [super dealloc];
}

- (BOOL)isPrivileged
{
    OSStatus err = 0;
    AuthorizationRights rights;
    AuthorizationItem items[1];

    // The authorization rights structure holds a reference to an array of AuthorizationItem structures that represent the rights for which you are requesting access.

    items[0].name = kAuthorizationRightExecute;
    items[0].value = (char *)[__launchPath fileSystemRepresentation];
    items[0].valueLength = strlen( items[0].value);
    items[0].flags = 0;

    rights.count=1;
    rights.items = items;

    // Since we specify kAuthorizationFlagExtendRights and have not specified kAuthorizationFlagInteractionAllowed, if the user isn't currently authorized to execute tools as root, we will fail to get the rights:
    err = AuthorizationCopyRights( __authorizationRef,&rights, kAuthorizationEmptyEnvironment, kAuthorizationFlagExtendRights, NULL);
    if (errAuthorizationSuccess != err) NSLog( @"isPrivileged: Authorization Framework failed with error: %d", err);

    return (errAuthorizationSuccess == err);
}

- (BOOL)setPrivileged: (BOOL)privileged
{
    if (privileged) {
	OSStatus err = 0;
	AuthorizationRights rights;
	AuthorizationItem items[1];

	if (!__authorizationRef) {
	    [self resetPrivileges];
	    if (!__authorizationRef) {
		NSLog( @"Repeated failure of acquisition of startup privileges");
		return NO;
	    }
	}

	// There should be one item in the AuthorizationItems array for each right you want to acquire.
	// The data in the value and valueLength is dependent on which right you want to acquire.
	// For the right to execute tools as root, kAuthorizationRightExecute, they should hold a pointer to a C string containing the path to the tool you want to execute, and the length of the C string path.
	// There needs to be one item for each tool you want to execute.

	items[0].name = kAuthorizationRightExecute; // which is a string and this is used to look in the policy database to see what authorization you want.
	items[0].value = (char *)[__launchPath fileSystemRepresentation];
	items[0].valueLength = strlen( items[0].value);
	items[0].flags = 0;

	rights.count=1;
	rights.items = items;

	// Since we specify kAuthorizationFlagExtendRights and have specified kAuthorizationFlagInteractionAllowed, if the user isn't currently authorized to execute tools as root, we will get an authentication panel:
	err = AuthorizationCopyRights (__authorizationRef, &rights, kAuthorizationEmptyEnvironment, (kAuthorizationFlagExtendRights|kAuthorizationFlagInteractionAllowed), NULL);
	if (errAuthorizationSuccess != err) {
	    NSLog( @"setPrivileged: Authorization Framework failed with error: %d", err);
	}
	return (errAuthorizationSuccess == err);
	// Could be NO on faulty user interaction
    }
    else {
	// We want to set this object from privileged to its default startup state. Note that this still leaves us more or less `authenticated' (no authentication required) if we were running as root in the first place
	[self resetPrivileges];
	return NO;
    }
}

- (AuthorizationRef)authorizationRef
{
    return __authorizationRef;    
}

- (void) reportRCSID
{
    NSLog( @"%s\n", rcsid);
}

@end

/*
 * $Log: GWAuthenticator.m,v $
 * Revision 1.2  2004/04/25 13:13:15  gctwnl
 * Forced commit of all .h and .m files of version 2.61.0
 *
 * Revision 1.1.1.1  2004/02/01 17:18:14  gctwnl
 * Initial import.
 *
 * Revision 3.3  2004/01/14 12:56:45  gerben
 * Version 2.39.0
 *
 * Revision 3.2  2002/12/10 14:24:48  gerben
 * Set sticky tag 3.2
 *
 * Revision 3.0  2002/12/10 14:04:22  gerben
 * Upping all files with a sticky tag of 1.99
 *
 * Revision 1.99  2002/12/10 13:52:49  gerben
 * Set all revision numbers to 1.99
 *
 * Revision 1.9  2002/10/21 20:16:54  gerben
 * Closing in on RC1 target
 *
 * Revision 1.8  2002/10/16 08:37:19  gerben
 * Changed log levels. Small fixes.
 *
 * Revision 1.7  2002/10/09 14:16:12  gerben
 * Changed as many %s/lossyCString uses to %@/fileSystemRepresentation/UTF8String
 *
 * Revision 1.6  2002/10/07 20:28:06  gerben
 * Removed circular CVS Log entry in C Comments at the end of files
 *
 * Revision 1.5  2002/10/03 13:38:04  gerben
 * Removed addChildWindow as child passes close to parent. Window can only close when there are no running tasks.
 *
 * Revision 1.4  2002/10/02 22:38:31  gerben
 * Added CVS Log in C Comment at the end of every file
 *
 */