//  Copyright (c) 2002-2004, Gerben Wierda, R&A
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
//  Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//  Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//
//  Neither the name of R&A nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#import "II2DocumentController.h"
#import "II2AppDelegate.h"
#import "II2Preferences.h"
#import "II2Package.h"
#import "II2Directory.h"
#import "GWSizeFormatter.h"

static char rcsid[] = "$Id: II2DocumentController.m,v 1.3 2004/10/15 14:13:06 gctwnl Exp $";

static NSDictionary *foregroundGetURLDictionary( NSString *fileURL)
{
    NSURL *url = [NSURL URLWithString:fileURL];
    NSData *urlData = urlDataFromURL( url);
    NSDictionary *dictionary;
    NSString *error;
    
    if (!urlData) {
	return nil;
    }
    dictionary = (NSDictionary *)CFPropertyListCreateFromXMLData( NULL, (CFDataRef)urlData, kCFPropertyListImmutable, (CFStringRef *)&error);
    
    return dictionary;
}

static NSString *foregroundGetURLString(NSString * fileName)
{
    NSURL *url = [NSURL URLWithString:fileName];
    NSData *urlData = urlDataFromURL( url);
    NSString *string;
    
    if (!urlData) {
	return nil;
    }
    string = [[[NSString alloc] initWithData:urlData encoding:NSUTF8StringEncoding] autorelease];
    
    return string;
}

@implementation II2DocumentController

- (BOOL)validateMenuItem:(NSMenuItem *)anItem
{
    // NSLog( @"%@ Validating %@", [[self class] description], [anItem description]);
    return YES;
}

- (IBAction)newDocument:(id)sender
{
    if (openURLPanel == nil)
    {
	// we need to load the nib
	if ([NSBundle loadNibNamed:@"II2OpenURLPanel" owner:self] == NO)
	{
	    NSRunAlertPanel(@"Error", @"Could not load OpenURLPanel.nib", @"Accept defeat", nil, nil);
	    return;
	}
    }
    [openURLPanel makeKeyAndOrderFront:self];
}

- (IBAction)openURLPanelOKPressed:(id)sender
{
    NSURL *url = [NSURL URLWithString:[urlTextField stringValue]];

    if (url) {
	if ((floor( NSAppKitVersionNumber) <= NSAppKitVersionNumber10_1)  && ([[url scheme] isEqualToString:@"ftp"])) {
	    NSRunAlertPanel(@"Error", @"FTP not supported by Apple Frameworks in Mac OS X Versions below 10.2", @"Use other scheme", nil, nil, nil);
	    return;
	}
	// close the window
	[openURLPanel performClose:self];
	[self openDocumentWithContentsOfURL:url display:YES];
    }
    else {
	NSRunAlertPanel(@"Error", @"Malformed URL", @"Try again", nil, nil, nil);
    }
}

- (IBAction)openURLPanelCancelPressed:(id)sender
{
    // close the window
    [openURLPanel performClose:self];
}

- (id)openDocumentWithContentsOfFile:(NSString *)fileName display:(BOOL)flag
{
    id pkgOrDir = [super openDocumentWithContentsOfFile:fileName display:flag];
    return pkgOrDir;
}


- (IBAction)openDocument:(id)sender
{
    int result;
    NSArray *fileTypes = [NSArray arrayWithObjects:@"ii2", @"iid", nil];
    NSOpenPanel *oPanel = [NSOpenPanel openPanel];

    [oPanel setAllowsMultipleSelection:YES];
    result = [oPanel runModalForTypes:fileTypes];
    if (result == NSOKButton) {
	NSArray *filesToOpen = [oPanel filenames];
	int i, count = [filesToOpen count];
	for (i=0; i<count; i++) {
	    NSString *aFile = [filesToOpen objectAtIndex:i];
	    [self openDocumentWithContentsOfFile:aFile display:YES];
	}
    }
}

- (id)openDocumentWithContentsOfURL:(NSURL *)anURL display:(BOOL)flag
{
    if ([anURL isFileURL]) {
	II2Package *filePackage = [self openDocumentWithContentsOfFile:[anURL path] display:flag];
	return filePackage;
    }
    else {
	if ([[anURL path] hasSuffix:@".iid"]) {
	    // A terrible hack, because we have a kind of non-document (remote IIDirectory without local file status) in our document architecture:
	    id newDoc = [self makeUntitledDocumentOfType:@"i-Package-v2-dir"];
	    [newDoc setStartURLString:[anURL absoluteString]];
	    [newDoc reload];
	    [self addDocument:newDoc];
	    [newDoc makeWindowControllers];
	    [[newDoc windowController] showWindow:nil];
	    [[[newDoc windowController] window] setTitle:[NSString stringWithFormat:@"%@ (%@)", [[anURL absoluteString] lastPathComponent], [anURL host]]];
	    return newDoc;
	}
	if (floor( NSAppKitVersionNumber) <= NSAppKitVersionNumber10_1) {
	    // Running on a 10.0 - 10.1.x system
	    if ([[anURL scheme] isEqualToString:@"ftp"]) {
		NSLog( @"FTP is not supported on Mac OS X version below 10.2 (and flaky at best in 10.2 and 10.3, so do not feel left out...)");
		return nil;
	    }
	}
	if (![ii2Prefs networkAccessDisabled]) {
	    II2Package *filePackage = nil;
	    BOOL isDir;
	    NSString *pkgDir = [[NSString stringWithFormat:@"%s/%s", [[ii2Prefs defaultSaveDir] fileSystemRepresentation], [[[anURL path] lastPathComponent] fileSystemRepresentation]] stringByExpandingTildeInPath];
	    NSString *plistFile = [NSString stringWithFormat:@"%s/%s.plist", [pkgDir fileSystemRepresentation], [[[pkgDir lastPathComponent] stringByDeletingPathExtension] fileSystemRepresentation]];
	    NSString *plistURLString = [NSString stringWithFormat:@"%@/%@.plist", [anURL absoluteString], [[[anURL path] lastPathComponent] stringByDeletingPathExtension]];
	    NSURL *plistURL = [NSURL URLWithString:plistURLString];
	    NSData *plistData = urlDataFromURL( plistURL);
	    
	    if (!plistData) {
		NSLog( @"Cannot read remote table of contents. The most probable cause is a download problem or a mistyped URL.");
		NSRunAlertPanel(@"Error", @"Cannot read remote table of contents. The most probable cause is a download problem.", @"Accept defeat", nil, nil);
		return nil;
	    }

	    if ([DEFFM fileExistsAtPath:[ii2Prefs defaultSaveDir] isDirectory:&isDir]) {
		if (!isDir) {
		    NSLog( [@"Default Save Directory problem. There is already a normal file of the same name at " stringByAppendingString:[ii2Prefs defaultSaveDir]]);
		    NSRunAlertPanel(@"Error", [@"Default Save Directory problem. There is already a normal file of the same name at " stringByAppendingString:[ii2Prefs defaultSaveDir]], @"Accept defeat", nil, nil);
		    return nil;
		}
	    }
	    else {
		if (![DEFFM createDirectoryAtPath:[ii2Prefs defaultSaveDir] attributes:nil]) {
		    NSLog( @"Cannot create default save directory %@. Permissions or directory problem?", [ii2Prefs defaultSaveDir]);
		    NSRunAlertPanel( @"Error", [NSString stringWithFormat:@"I am unable to create the default save directory %@ on disk. Permissions or directory problem?", [ii2Prefs defaultSaveDir]], @"Accept", nil, nil, nil);
		    return nil;
		}
	    }
	    
	    if ([DEFFM fileExistsAtPath:pkgDir isDirectory:&isDir]) {
		if (!isDir) {
		    NSLog( [@"There is already a normal file of the same name at " stringByAppendingString:pkgDir]);
		    NSRunAlertPanel(@"Error", [@"There is already a normal file of the same name at " stringByAppendingString:pkgDir], @"Accept defeat", nil, nil);
		    return nil;
		}
		// This one is in fact too early. If we are opening from a new URL and the one on disk has an old URL, the old URL is used initially. This can't be helped because we in fact do not know the URL of the package on disk until it is opened.
		filePackage = [self openDocumentWithContentsOfFile:pkgDir display:flag];
		if (filePackage) {
		    if (![[filePackage urlString] isEqualToString:[anURL absoluteString]]) {
			[filePackage setURLString:[anURL absoluteString]];
			[filePackage readRemotePkgPlist:YES]; // Re-read remote as the URL changed
			[[filePackage windowController] updatePropertiesDisplay];
			NSBeginAlertSheet( @"Warning", @"OK", nil, nil, [[filePackage windowController] window], nil, nil, nil, nil, @"The URL of the remote package you selected differs from the one of existing package on Disk, but the package names are equal so it is assumed they represent the same software. The package on disk has therefore been opened instead. The new URL has been set in Properties and your are advised to update from the new URL immediately. If you do not want this new URL, choose Reset To Default Properties from the i-Package menu to reset both URL and Install Location to package defaults.");
		    }
		}
		return filePackage;
	    }
	    if (![DEFFM createDirectoryAtPath:pkgDir attributes:nil]) {
		NSLog( @"Cannot create local package bundle");
		NSRunAlertPanel( @"Error", [NSString stringWithFormat:@"I am unable to create the package .ii2 bundle %@ on disk. Permissions or directory problem?", pkgDir], @"Accept", nil, nil, nil);
		return nil;
	    }
	    if ([plistData writeToFile:plistFile atomically:NO]) {
		NSString *md5URLString = [NSString stringWithFormat:@"%@/%@.plist.md5", [anURL absoluteString], [[[anURL path] lastPathComponent] stringByDeletingPathExtension]];
		NSString *customplistname = [pkgDir stringByAppendingString:@"/Custom.plist"];
		NSDictionary *customdict = [NSDictionary dictionaryWithObject:[anURL absoluteString] forKey:@"default_url"];
		NSString *sigFile = [NSString stringWithFormat:@"%s/%s.plist.sig", [pkgDir fileSystemRepresentation], [[[pkgDir lastPathComponent] stringByDeletingPathExtension] fileSystemRepresentation]];
		NSString *sigURLString = [NSString stringWithFormat:@"%@/%@.plist.sig", [anURL absoluteString], [[[anURL path] lastPathComponent] stringByDeletingPathExtension]];
		NSURL *sigURL = [NSURL URLWithString:sigURLString];
		NSData *sigData = urlDataFromURL( sigURL);
		
		[customdict writeToFile:customplistname atomically:NO];
		
		if (sigData) {
		    if (![sigData writeToFile:sigFile atomically:NO]) {
			if ([ii2Prefs mainLogLevel] > 0) NSLog( @"Unable to write sig file %@. Sig %@ probably unavailable.", sigFile, sigURLString);
		    }
		}
		
		if (checkMD5OfFileAgainstContentsOfURL( plistFile, md5URLString)) {
		    filePackage = [self openDocumentWithContentsOfFile:pkgDir display:flag];
		}
		else {
		    if ([ii2Prefs mainLogLevel] > 0) NSLog( @"MD5 checksum for %@ is not equal to contents of %@", plistFile, md5URLString);
		    if ([ii2Prefs keepInvalidMd5]) {
			filePackage = [self openDocumentWithContentsOfFile:pkgDir display:flag];
		    }
		    else {
			NSRunAlertPanel(@"Error", @"MD5 Checksum of i-Package Table of Contents is not equal to contents of remote MD5 Checksum file", @"Give Up", nil, nil, nil);
			[DEFFM removeFileAtPath:pkgDir handler:nil];
		    }
		}
	    }
	    return filePackage;
	}
	else {
	    NSRunAlertPanel( @"Error", @"It is not possible to open remote packages when network access has been disabled in your preferences.", @"Accept", nil, nil, nil);
	}
    }
    return nil;
}

- (void)openUpdatedLocalPackages:(NSString *)inDirectory
{
    int i;
    NSArray *fileArray = [DEFFM directoryContentsAtPath:[ii2Prefs defaultSaveDir]];
    
    if ([ii2Prefs networkAccessDisabled]) {
	NSLog( @"Cannot check remote packages when network setting is set to disabled in i-Installer preferences");
	NSRunAlertPanel( @"Error", @"It is not possible to open remote packages when network access has been disabled in your preferences.", @"Accept", nil, nil, nil);
	return;
    }
    
    for (i=0; i<[fileArray count]; i++) {
	NSString *pkgName = [NSString stringWithFormat:@"%@/%@", [ii2Prefs defaultSaveDir], [fileArray objectAtIndex:i]];
	NSString *plistName, *customPlistName, *pkgBaseName;
	NSDictionary *plistDict, *customPlistDict, *remotePlistDict;
	NSString *remoteLocationString, *remotePlistURLString, *remoteUpdateStampString, *remoteUpdateStampContents;
	NSNumber *localTimestamp, *remoteTimestamp;
	
	if (![pkgName hasSuffix:@".ii2"]) {
	    if ([ii2Prefs mainLogLevel] > 8) NSLog( @"%@ is not an i-Package. Wrong extension. Skipping...", pkgName);
	    continue;
	}
	pkgBaseName = [[[pkgName lastPathComponent] stringByDeletingPathExtension] retain];
	plistName = [NSString stringWithFormat:@"%@/%@.plist", pkgName, pkgBaseName];
	plistDict = [[NSDictionary dictionaryWithContentsOfFile:plistName] retain];
	if (!plistDict) {
	    if ([ii2Prefs mainLogLevel] > 0) NSLog( @"%@ not found. Missing package or missing Table of Contents. Skipping...", plistName);
	    continue;
	}
	customPlistName = [[pkgName stringByAppendingString:@"/Custom.plist"] retain];
	customPlistDict = [[NSDictionary dictionaryWithContentsOfFile:customPlistName] retain];
	
	remoteLocationString = [customPlistDict objectForKey:@"default_url"];
	if (!remoteLocationString) {
	    remoteLocationString = [plistDict objectForKey:@"default_url"];
	}
	if (!remoteLocationString) {
	    if ([ii2Prefs mainLogLevel] > 0) NSLog( @"%@ is not an i-Package. Missing URL in Table of Contents. Skipping...", pkgName);
	    continue;
	}
	remoteUpdateStampString = [NSString stringWithFormat:@"%@/%@.startupdate", remoteLocationString, pkgBaseName];
	remoteUpdateStampContents = foregroundGetURLString( remoteUpdateStampString);
	if (remoteUpdateStampContents) {
	    NSLog( @"The remote location for %@ is currently being updated\n", [pkgName stringByAbbreviatingWithTildeInPath]);
	    continue;
	}
	
	remotePlistURLString = [NSString stringWithFormat:@"%@/%@.plist", remoteLocationString, pkgBaseName];
	remotePlistDict = foregroundGetURLDictionary( remotePlistURLString);
	if (!remotePlistDict) {
	    if ([ii2Prefs mainLogLevel] > 0) NSLog( @"%@ is not an i-Package. Missing Table of Contents. Skipping...", remoteLocationString);
	    continue;
	}
	localTimestamp = [NSNumber numberWithLong:[[plistDict objectForKey:@"timestamp"] intValue]];
	remoteTimestamp = [NSNumber numberWithLong:[[remotePlistDict objectForKey:@"timestamp"] intValue]];
	if ([localTimestamp compare:remoteTimestamp] != NSOrderedSame) {
	    NSString *timediff = ([localTimestamp compare:remoteTimestamp] == NSOrderedAscending ? @"newer" :@"older");
	    unsigned int changedSize = 0;
	    NSString *formattedSize;
	    NSDictionary *localFilesDict = [plistDict objectForKey:@"files"];
	    NSDictionary *remoteFilesDict = [remotePlistDict objectForKey:@"files"];
	    
	    NSEnumerator *enumerator = [remoteFilesDict keyEnumerator];
	    NSString *remoteFileKey;
	    
	    NSLog( @"The remote location for %@ has a different version\n", [pkgName stringByAbbreviatingWithTildeInPath]);

	    while ((remoteFileKey = [enumerator nextObject])) {
		NSDictionary *localFileDict = [localFilesDict objectForKey:remoteFileKey];
		if (!localFileDict) {
		    // New file in package
		    changedSize += [[[remoteFilesDict objectForKey:remoteFileKey] objectForKey:@"size"] intValue];
		}
		else {
		    // File in both local and remote
		    NSString *localmd5 = [localFileDict objectForKey:@"md5"];
		    NSString *remotemd5 = [[remoteFilesDict objectForKey:remoteFileKey] objectForKey:@"md5"];
		    if (![localmd5 isEqualToString:remotemd5]) {
			changedSize += [[[remoteFilesDict objectForKey:remoteFileKey] objectForKey:@"size"] intValue];
		    }
		}
	    }
	    formattedSize = [[GWSizeFormatter sharedFormatter] stringForObjectValue:[NSNumber numberWithUnsignedInt:changedSize]];
	    
	    if ([ii2Prefs mainLogLevel] > 1) NSLog( @"There is a %@ version available for %@ (%@ changed)\n", timediff, [pkgName stringByAbbreviatingWithTildeInPath], formattedSize);
	    {
		II2Package *package = (II2Package *)[[NSDocumentController sharedDocumentController] openDocumentWithContentsOfFile:pkgName display:YES];
		II2WindowController *wC = [package windowController];
		if (wC) {
		    [[wC messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"%@ of this package has been changed in a remotely available update\n", formattedSize] attributes:[self normalAttributes]] autorelease]];
		}
	    }
	}
    }
    NSRunAlertPanel( @"Finished Checking for Updates", @"All i-Packages files in your local default save directory have been checked.", @"OK", nil, nil, nil);
    return;
}

- (void)glupTaskDidFinish: (GLUPTaskRunner *)task withSuccess:(BOOL)success
{

}

- (void)glupTaskWillStart: (GLUPTaskRunner *)task
{
    
}

- (void) reportRCSID
{
    NSLog( @"%s\n", rcsid);
}

@end

/*
 * $Log: II2DocumentController.m,v $
 * Revision 1.3  2004/10/15 14:13:06  gctwnl
 * 2.68
 *
 * Revision 1.2  2004/04/25 13:13:21  gctwnl
 * Forced commit of all .h and .m files of version 2.61.0
 *
 * Revision 1.1.1.1  2004/02/01 17:18:21  gctwnl
 * Initial import.
 *
 * Revision 3.9  2004/01/14 12:56:46  gerben
 * Version 2.39.0
 *
 * Revision 3.8  2004/01/05 15:31:34  gerben
 * *** empty log message ***
 *
 * Revision 3.7  2003/11/20 11:56:06  gerben
 * 2.32.0 work in progress
 *
 * Revision 3.6  2003/05/18 13:35:17  gerben
 * Version 2.21.0. Improved i-Directory handling.
 *
 * Revision 3.5  2003/04/02 23:19:53  gerben
 * Creates default save directory if it does not exist yet.
 *
 * Revision 3.4  2003/03/11 22:23:33  gerben
 * Added warning when package cannot be created. Removed p flag from gnutar, unarchived files will be owned by root or user depending on authentication
 *
 * Revision 3.3  2003/02/06 21:04:13  gerben
 * Version 2.3.0
 *
 * Revision 3.2  2002/12/10 14:24:50  gerben
 * Set sticky tag 3.2
 *
 * Revision 3.0  2002/12/10 14:04:23  gerben
 * Upping all files with a sticky tag of 1.99
 *
 * Revision 1.99  2002/12/10 13:52:51  gerben
 * Set all revision numbers to 1.99
 *
 * Revision 1.29  2002/11/22 23:54:47  gerben
 * RC4
 *
 * Revision 1.28  2002/10/28 21:49:45  gerben
 * A lot of changes wrt known packages.
 *
 * Revision 1.27  2002/10/27 11:19:50  gerben
 * Sig checking changed from remote to local for TOC. Update does not update when timestamps inside TOCs are equal.
 *
 * Revision 1.26  2002/10/25 13:47:58  gerben
 * New Known Packages Window
 *
 * Revision 1.25  2002/10/24 20:18:17  gerben
 * Better sheet
 *
 * Revision 1.24  2002/10/24 20:12:31  gerben
 * Better handling of package name clashes and changed properties
 *
 * Revision 1.23  2002/10/24 13:03:31  gerben
 * Disable menu items that can affect downloads and availability of files when package is active.
 *
 * Revision 1.22  2002/10/21 20:16:52  gerben
 * Closing in on RC1 target
 *
 * Revision 1.21  2002/10/17 18:48:58  gerben
 * Cache/NoCache handling on downloads. Some other changes.
 *
 * Revision 1.20  2002/10/17 10:18:59  gerben
 * Splash screen. MD5 check on opening remote package.
 *
 * Revision 1.19  2002/10/15 21:14:21  gerben
 * A less klunky known packages menu
 *
 * Revision 1.18  2002/10/15 14:31:43  gerben
 * Last fixes. No-network access. Later creation of downloaded file.
 *
 * Revision 1.17  2002/10/13 22:22:26  gerben
 * Re-created file menu so validating menu items work. Changed code for inspect and fatten to use generic task array starter.
 *
 * Revision 1.16  2002/10/13 21:20:14  gerben
 * Fixed a serious bug where the task rewrite was a bit release-happy
 *
 * Revision 1.15  2002/10/09 14:16:11  gerben
 * Changed as many %s/lossyCString uses to %@/fileSystemRepresentation/UTF8String
 *
 * Revision 1.14  2002/10/08 21:34:27  gerben
 * Working GPG signature checks, shared GPG cache for II2.
 *
 * Revision 1.13  2002/10/08 10:03:42  gerben
 * Several small changes, amongst others first responder behaviour.
 *
 * Revision 1.12  2002/10/07 20:51:30  gerben
 * NSProgressIndicator on package window. Some renaming of functions.
 *
 * Revision 1.11  2002/10/07 20:28:05  gerben
 * Removed circular CVS Log entry in C Comments at the end of files
 *
 * Revision 1.10  2002/10/07 19:08:59  gerben
 * Added checking of external MD5's (either URL or file based).
 *
 * Revision 1.9  2002/10/07 11:52:27  gerben
 * Several fixes related to possible invalid content.
 *
 * Revision 1.8  2002/10/03 13:38:03  gerben
 * Removed addChildWindow as child passes close to parent. Window can only close when there are no running tasks.
 *
 * Revision 1.7  2002/10/02 22:38:30  gerben
 * Added CVS Log in C Comment at the end of every file
 */