//  Copyright (c) 2002-2004, Gerben Wierda, R&A
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
//  Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//  Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//
//  Neither the name of R&A nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#import <Cocoa/Cocoa.h>
#import "II2TaskSet.h"
#import "II2DownloadTask.h"
#import "II2ActionTask.h"
#import "GLUPTaskRunner.h"
#import "II2WindowController.h"
#import "GWAuthenticator.h"

// NSOrderedSame etc are -1,0,1. For timestamps there is also the unordered state when the remote package is being updated or not found.
typedef enum {
    II2IsUnavailable = -3,
    II2IsBeingUpdated = -2,
    II2OrderedAscending = NSOrderedAscending,
    II2OrderedSame,
    II2OrderedDescending
} II2TimestampComparisonResult;

extern id ii2TagNameArray[];

@interface II2Package : NSDocument
{
    NSDictionary 	*__pkgPlist;		// As distrbuted by package provider
    NSMutableDictionary	*__customPkgPlist;      // Settings the user changed
    NSDictionary 	*__remotePkgPlist;      // As found in a remote repository
    NSMutableDictionary *__installedPkgPlist;   // As found in the install location
    NSMutableDictionary *__md5Checks;
    NSMutableDictionary *__fileDataTypes;
    NSMutableDictionary *__sigChecks;
    NSMutableArray	*__fileTableSource;
    NSMutableArray	*__completeFileArray;

    NSString	*__baseName;			// foo part of /some/path/foo.ii2
    NSString	*__dirName;			// /some/path/foo.ii2
    NSString	*__dirNameWithSeparator;	// /some/path/foo.ii2/ for convenience
    NSString	*__urlNameWithSeparator;	// http://foo.bar/path/blabla.ii2/ for convenience
    
    // Process handling
    GWAuthenticator *__auth;
    II2TaskSet	*__ii2RunningTaskSet;

    // We only have one window controller in this app for each doc. We cache it to simplify coding.
    II2WindowController *__cachedWindowController;
    
    // We have a dictionary for environment settings we can set in one task and pass on to another. 
    NSMutableDictionary *__environment;
}


- (id)initWithContentsOfFile:(NSString *)fileName ofType:(NSString *)docType;
- (II2WindowController *)windowController;

@end

@interface II2Package (II2PackageInfo)
// Part number from file name
- (int) partNumberFromArchivePartFileName: (NSString *)archivePartname;

// file wrapper contents
- (void)returnToDefaults;

// Miscellaneous
- (NSString *)baseName;
- (NSString *)packageName;
- (unsigned int)totalSize;
- (unsigned int)sizeOnDisk;
- (BOOL)isInteractive;
- (BOOL)needsResourceForksFixup;
- (BOOL)hasNonStandardFiles;
- (BOOL)hasArchive;
- (BOOL)hasRemoveScript;
- (BOOL)hasConfigureScript;
- (BOOL)hasSelectorScript;

- (BOOL)hasTabViewSettings;
- (BOOL)hideTabForTag:(II2TabTagType)tag;
- (void)setHideTabForTag:(II2TabTagType)tag toState:(BOOL)state;

// Authentication
- (BOOL)needsAuthentication;
- (void)setAuthenticationRequired:(BOOL)required;

// Locking
- (BOOL)isLocked;
- (void)setLockStatus:(BOOL)locked;

// Writable
- (BOOL)isWritable;

// Can be fattened: unlocked, network not shut down in preferences and ona writable filesystem
- (BOOL)canBeFattened;

// Remote URL
- (NSString *)urlString;
- (void)setURLString:(NSString *)newURLString;

// Install location
- (NSString *)installLocation;
- (void)setInstallLocation: (NSString*)location;
- (BOOL)isRelocatable;

// Files in package
- (NSString *)fullPathForFileName:(NSString *)fileName;
- (NSString *)fullURLPathForFileName:(NSString *)fileName;
- (unsigned int)sizeOfFile:(NSString *)fileName;
- (BOOL)fileSetIsAvailable:(NSString *)fileName;
- (BOOL)fileIsAvailable:(NSString *)fileName;
- (BOOL)fileSetHasValidMd5:(NSString *)fileName;
- (BOOL)fileHasValidMd5:(NSString *)fileName;
- (BOOL)fileSetHasValidSig:(NSString *)fileName;
- (BOOL)fileHasValidSig:(NSString *)fileName;
- (NSString *)fileNameForInspect:(NSString *)fileName withModifiers:(unsigned int)modifiers;
- (BOOL)packageHasMd5ForFileSet:(NSString *)fileName;
- (BOOL)packageHasMd5ForFile:(NSString *)fileName;
- (BOOL)packageHasSigForFileSet:(NSString *)fileName;
- (BOOL)packageHasSigForFile:(NSString *)fileName;
- (BOOL)fileIsPartOfPackage:(NSString *)fileName;
- (void)clearChecksForFile:(NSString *)fileName;
- (void)clearChecksForFileSet:(NSString *)fileName;
- (NSString *)fileTypeDisplayString:(NSString *)fileName;
- (NSString *)fileNameAtIndex:(int)index;
- (NSMutableArray *)allocFilePartsArrayFromFileName:(NSString *)fileName filter:(BOOL)filter;
- (NSArray *)dependentFilesForStage:(NSString *)stage filter:(BOOL)filter;

// Remote plist
- (BOOL)readRemotePkgPlist: (BOOL)force;

// Installed plist
- (BOOL)readInstalledPkgPlist:(BOOL)force;

// Versions
- (NSNumber *)pkgIIVersion;
- (NSNumber *)remoteIIVersion;

// Installed package NSDictionary
- (NSDictionary *)installedPackageDictionary;

// Timestamps1
- (NSString *)pkgMinII2stamp;
- (NSString *)pkgMinII2stampDisplay;
- (NSString *)pkgTimestamp;
- (NSString *)pkgTimestampDisplay;
- (NSString *)installedTimestamp;
- (NSString *)installedTimestampDisplay;
- (NSString *)remoteTimestamp;
- (NSString *)remoteStatusDisplay;
- (NSString *)remoteUpdateTimestamp;
- (II2TimestampComparisonResult)remoteStatus;

// Foreground downloading
- (NSString *)foregroundGetURLString: (NSString *)fileName;
- (NSDictionary *)foregroundGetURLDictionary: (NSString *)fileName;

// Data from file
- (NSString *)stringFromFileContents:(NSString *)fileName;
- (NSData *)dataFromFileContents:(NSString *)fileName;

// Additions to environment for configure/remove subtasks
- (void)setEnvironmentValue:(NSString *)value forKey:(NSString *)key;
- (void)unsetEnvironmentValueforKey:(NSString *)key;
- (void)resetEnvironment;

// Set selector set to be used
- (void)setSelectorSets:(NSArray *)setArray;
- (NSArray *)selectedSelectorSets;
- (NSArray *)selectedKnownSets;
- (BOOL)setIdInAvailableSets: (NSString *)knownSetId;
- (BOOL)setIdInSelectedSets: (NSString *)knownSetId;
- (BOOL)setIdInConfiguredSets: (NSString *)knownSetId;
- (BOOL)setIdInInstalledSets: (NSString *)knownSetId;
// Store what has been configured/installed wrt sets
- (void)setSelectorSet:(NSString *)setId configured:(BOOL)configured;
- (void)setSelectorSet:(NSString *)setId installed:(BOOL)installed;
- (NSArray *)availableSelectorSets;
- (NSArray *)configuredSelectorSets;
- (NSArray *)installedSelectorSets;
// Extended selector set array which can be passed to scripts via Interaction Manager
- (NSArray *)extendedSelectorSetArray;

// Is archive part number in one of the sets?
- (BOOL)intInSelectedSets:(int)archivePartNumber;

// Receipts file
- (void)writeReceipt;
- (void)removeReceipt;
@end

@interface II2Package (II2PackageUserCommunication)
// Sheet handling
- (void)unimplementedSheet;
- (void)progressSheet;
- (void)doUpdateCheckAndAction:(NSString *)actionString withObject:(id)object;
- (void)compareSheet:(II2TimestampComparisonResult)compare forAction:(NSString *)actionString withObject:(id)object;
- (void)endSheetWithDictionary: (NSDictionary *)dictionary;
- (BOOL)reportPkgDependencies;
- (void)endSheetWithDictionaryForDependency: (NSDictionary *)dictionary;
@end

@interface II2Package (II2PackageTableSource)
// tableSources
- (int)numberOfRowsInTableView:(NSTableView *)aTableView;
- (id)tableView:(NSTableView *)aTableView objectValueForTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex;
- (void)setupFileTableSourceAndInfo;
- (id)taskDataSource;
@end

@interface II2Package (II2PackageActions) <II2DownloadTaskOwnerProtocol, II2ActionTaskOwnerProtocol, GLUPTaskOwnerProtocol, II2TaskSetHandlerProtocol>
- (void)setCommandEnvironment:(GLUPTaskRunner *)task withUserDefined:(BOOL)userDefined;
- (NSArray *)allocTasksForFileSetAvailable:(NSString *)fileName forTask:(id <II2TaskDependencyProtocol>)dependentTask filter:(BOOL)filter;
- (BOOL)checkForDownloadsInTaskArray:(NSArray *)tasks;
- (BOOL)prepareInstallDir;
- (NSArray *)allocTasksForAction:(NSString *)action;
- (void)destroyTaskArray:(NSArray *)tasks;
- (void)markTaskArray:(NSArray *)tasks forAction:(NSString *)action;
- (void)doActionTasks:(NSArray *)tasks;
- (void)action:(NSString *)action;
- (BOOL)update;
- (void)inspect:(NSString *)fileName withModifiers:(unsigned int)modifiers inView:(NSTextView *)aView;
- (void)showInfo; // Convenience method
- (BOOL)deleteFile:(NSString *)fileName;
- (void)gpgCheckToC;
- (GLUPTaskRunner *)md5CheckTask:(NSString *)fileName forTask:(id <II2TaskDependencyProtocol>)dependentTask;
- (NSArray *)gpgCheckTasks:(NSString *)fileName forTask:(id <II2TaskDependencyProtocol>)dependentTask;
- (NSArray *)dataTypeCheckTasks:(NSString *)fileName forTask:(id <II2TaskDependencyProtocol>)dependentTask;
- (void)glupTaskDidFinish: (GLUPTaskRunner *)task withSuccess:(BOOL)success;
- (void)glupTaskWillStart: (GLUPTaskRunner *)task;
- (void)downloadTaskDidFinish: (II2DownloadTask *)task withSuccess:(BOOL)success;
- (void)downloadTaskWillStart: (II2DownloadTask *)task;
- (void)actionTaskDidFinish: (II2ActionTask *)task withSuccess:(BOOL)success;
- (void)actionTaskWillStart: (II2ActionTask *)task;
- (BOOL)ii2TaskObject: (id)object1 isEqualTo: (id)object2;
- (void)abortTaskAtIndex:(int)index;
- (void)fatten;
- (void)thin;
- (BOOL)active;
@end

extern NSString *displayStringFromTimestampString( NSString *stamp);

extern BOOL md5StringsAreEqual( NSString *md5String1, NSString *md5String2);
extern BOOL checkMD5OfFileAgainstCompanionMD5File( NSString *fullPath);
extern BOOL checkMD5OfFileAgainstContentsOfURL( NSString *fullPath, NSString *fullMD5URLString);
extern BOOL checkMD5OfFileAgainstContentsOfFile( NSString *fullPath, NSString *md5FullPath);

extern BOOL intInArray( int number, NSArray *array);

/*
 * $Log: II2Package.h,v $
 * Revision 1.4  2004/04/30 12:18:06  gctwnl
 * 2.62.0
 *
 * Revision 1.3  2004/04/25 13:13:21  gctwnl
 * Forced commit of all .h and .m files of version 2.61.0
 *
 * Revision 1.2  2004/03/21 16:52:21  gctwnl
 * 2.58.0
 *
 * Revision 1.1.1.1  2004/02/01 17:18:22  gctwnl
 * Initial import.
 *
 * Revision 3.12  2004/01/27 22:49:32  gerben
 * 2.45.0
 *
 * Revision 3.11  2004/01/27 12:57:39  gerben
 * 2.44.0
 *
 * Revision 3.10  2004/01/23 00:31:16  gerben
 * 2.42.0 release
 *
 * Revision 3.9  2004/01/14 12:56:46  gerben
 * Version 2.39.0
 *
 * Revision 3.8  2004/01/05 15:31:34  gerben
 * *** empty log message ***
 *
 * Revision 3.7  2003/11/09 21:43:06  gerben
 * 2.32.0
 *
 * Revision 3.6  2003/11/08 23:38:01  gerben
 * 2.30.0 release candidate
 *
 * Revision 3.5  2003/09/18 22:40:54  gerben
 * 2.26.0 Fragmented archives
 *
 * Revision 3.4  2003/02/23 23:29:11  gerben
 * Package Locking
 *
 * Revision 3.3  2002/12/16 13:13:37  gerben
 * Bug fix: The install directory was not created when one chose to install while ignoring the remote availability of another version.
 *
 * Revision 3.2  2002/12/10 14:24:52  gerben
 * Set sticky tag 3.2
 *
 * Revision 3.0  2002/12/10 14:04:23  gerben
 * Upping all files with a sticky tag of 1.99
 *
 * Revision 1.99  2002/12/10 13:52:51  gerben
 * Set all revision numbers to 1.99
 *
 * Revision 1.40  2002/12/10 08:55:24  gerben
 * II2Package code cleanup, remove of cached view pointers. Added relocatable handling in package dependencies.
 *
 * Revision 1.39  2002/12/09 15:08:59  gerben
 * Report printing
 *
 * Revision 1.38  2002/12/02 23:24:40  gerben
 * Basic dependency handling
 *
 * Revision 1.37  2002/11/30 10:05:53  gerben
 * Rearranged code for II2Package to use catagories instead of code inclusion.
 *
 * Revision 1.36  2002/11/24 23:09:27  gerben
 * Tom's changes and remarks mainly
 *
 * Revision 1.35  2002/10/31 14:26:56  gerben
 * Moved authentication setting to package window
 *
 * Revision 1.34  2002/10/30 21:42:22  gerben
 * RC3a
 *
 * Revision 1.33  2002/10/29 09:36:22  gerben
 * RC2
 *
 * Revision 1.32  2002/10/28 21:49:45  gerben
 * A lot of changes wrt known packages.
 *
 * Revision 1.31  2002/10/27 12:38:34  gerben
 * Lazy update checking. Update followed by action for inspect and fatten.
 *
 * Revision 1.30  2002/10/25 20:36:11  gerben
 * Layout changes.
 *
 * Revision 1.29  2002/10/25 19:52:35  gerben
 * Changes to update reporting.
 *
 * Revision 1.28  2002/10/24 20:12:32  gerben
 * Better handling of package name clashes and changed properties
 *
 * Revision 1.27  2002/10/21 20:16:52  gerben
 * Closing in on RC1 target
 *
 * Revision 1.26  2002/10/16 21:28:21  gerben
 * A complete new rewrite of taskss, processes and stuff.
 *
 * Revision 1.24  2002/10/14 12:34:19  gerben
 * Removed dead code
 *
 * Revision 1.23  2002/10/12 10:54:38  gerben
 * A lot of changes, mostly when producing output/feedback.
 *
 * Revision 1.22  2002/10/10 00:51:24  gerben
 * Many changes. Committing a stable state.
 *
 * Revision 1.21  2002/10/08 21:34:28  gerben
 * Working GPG signature checks, shared GPG cache for II2.
 *
 * Revision 1.20  2002/10/07 20:51:30  gerben
 * NSProgressIndicator on package window. Some renaming of functions.
 *
 * Revision 1.19  2002/10/07 20:28:05  gerben
 * Removed circular CVS Log entry in C Comments at the end of files
 *
 * Revision 1.18  2002/10/07 20:20:03  gerben
 * More md5 stuff and testing of spinning disk progressindicator stuff.
 *
 * Revision 1.17  2002/10/07 19:08:59  gerben
 * Added checking of external MD5's (either URL or file based).
 *
 * Revision 1.16  2002/10/07 11:52:27  gerben
 * Several fixes related to possible invalid content.
 *
 * Revision 1.15  2002/10/03 13:38:03  gerben
 * Removed addChildWindow as child passes close to parent. Window can only close when there are no running tasks.
 *
 * Revision 1.14  2002/10/02 22:38:30  gerben
 * Added CVS Log in C Comment at the end of every file
 *
 */