//  Copyright (c) 2002-2004, Gerben Wierda, R&A
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
//  Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//  Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//
//  Neither the name of R&A nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <sys/types.h>
#include <unistd.h>

#import "II2AppDelegate.h"
#import "II2Preferences.h"
#import "II2Package.h"
#import "II2InteractionManager.h"
#import "GWSizeFormatter.h"
#import "II2ToolFinder.h"

static char rcsidActions[] = "$Id: II2PackageActions.m,v 1.11 2004/10/11 20:16:24 gctwnl Exp $";

static int compareFileDictOnSize( id first, id second, void *context)
{
    unsigned int firstSize = [[first objectForKey:@"size"] intValue];
    unsigned int secondSize = [[second objectForKey:@"size"] intValue];
    if (firstSize < secondSize) return NSOrderedDescending;
    else if (firstSize == secondSize) return NSOrderedSame;
    else return NSOrderedAscending;
}

@implementation II2Package (II2PackageActions)

// Implementation Note: tasks and task arrays are in principle free-floating objects and when tasks end they release themselves. Hence, their creation cannot include autorelease and the tasks are allocated but not released. Probably it would in the end have been simpler to do the classic approach of controlled ownership in this application. Oh well, we never stop learning....

-(NSMutableArray *)allocTasksForFileAvailable:(NSString *)fileName forTask:(id <II2TaskDependencyProtocol>)dependentTask
{
    GLUPTaskRunner *md5Task = nil;
    NSArray *fileTypeTasks = nil, *sigCheckTasks = nil;
    II2DownloadTask *downloadTask = nil;
    NSMutableArray *tasksToRun = [[NSMutableArray alloc] init];

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: allocTasksForFileAvailable %@", fileName);
    if (![self fileIsAvailable:fileName]) {
	[self clearChecksForFile:fileName];
	downloadTask = [[II2DownloadTask alloc] initWithOwner:self urlString:[self fullURLPathForFileName:fileName] forFile:[self fullPathForFileName:fileName] ofSize:[self sizeOfFile:fileName]];
	if (!downloadTask) {
	    NSLog( @"Failed to create download task for file availability");
	    return nil;
	}
    }
    if (![self fileHasValidMd5:fileName]) {
	md5Task = [self md5CheckTask:fileName forTask:dependentTask];
    }
    if (![__fileDataTypes objectForKey:fileName]) {
	fileTypeTasks = [self dataTypeCheckTasks:fileName forTask:dependentTask];
	if (md5Task) {
	    [md5Task addDependentTasks:fileTypeTasks];
	}
    }
    if (![self fileHasValidSig:fileName] && [ii2Prefs useGPG]) {
	sigCheckTasks = [self gpgCheckTasks:fileName forTask:dependentTask];
    }
    if (downloadTask) {
	[downloadTask setMessageView:[__cachedWindowController messageView]];
	[downloadTask addDependentTask:md5Task];
	if (sigCheckTasks) {
	    [downloadTask addDependentTask:[sigCheckTasks objectAtIndex:1]];
	    [downloadTask addDependentTask:[sigCheckTasks objectAtIndex:0]];
	}
	[downloadTask addDependentTask:[fileTypeTasks objectAtIndex:1]];
	[downloadTask addDependentTask:[fileTypeTasks objectAtIndex:0]];
	[tasksToRun addObject:downloadTask];
    }
    if (md5Task) {
	[tasksToRun addObject:md5Task];
    }
    if (sigCheckTasks) {
	[tasksToRun addObject:[sigCheckTasks objectAtIndex:0]];
	[tasksToRun addObject:[sigCheckTasks objectAtIndex:1]];
    }
    if (fileTypeTasks) {
	[tasksToRun addObject:[fileTypeTasks objectAtIndex:0]];
	[tasksToRun addObject:[fileTypeTasks objectAtIndex:1]];
    }
    return tasksToRun;
}

-(NSArray *)allocTasksForFileSetAvailable:(NSString *)fileName forTask:(id <II2TaskDependencyProtocol>)dependentTask filter:(BOOL)filter
{
    NSArray *partSet = [self allocFilePartsArrayFromFileName:fileName filter:filter];
    int i;
    NSMutableArray *tasksToRun = [[NSMutableArray alloc] init];

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: allocTasksForFileSetAvailable %@", fileName);
    for (i = 0; i < [partSet count]; i++) {
	NSArray *newTasks = [self allocTasksForFileAvailable:[partSet objectAtIndex:i] forTask:dependentTask];
	if (newTasks) {
	    if ([newTasks count]) {
		if ([tasksToRun lastObject]) {
		    [(id <II2TaskDependencyProtocol>)[tasksToRun lastObject] addDependentTask:(id <II2TaskDependencyProtocol>)[newTasks objectAtIndex:0]];
		}
		[tasksToRun addObjectsFromArray:newTasks];
	    }
	    [newTasks release];
	}
    }
    [partSet release];
    return tasksToRun;
}

- (NSArray *)__allocFattenTasks
{
    int i, max = [__completeFileArray count];
    NSMutableArray *allTasks = [[NSMutableArray alloc] init];
    
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: __allocFattenTasks");
    for (i=0; i<max; i++) {
	NSString *fileName = [[__completeFileArray objectAtIndex:i] objectForKey:@"file"];
	NSArray *tasksToLaunch = [self allocTasksForFileAvailable:fileName forTask:nil];
	if (!tasksToLaunch) {
	    NSLog( @"Some weird task allocation failure ocurred while creeting fattening tasks.");
	    return nil;
	}
	if ([tasksToLaunch count]) {
	    [allTasks addObjectsFromArray:tasksToLaunch];
	}
	[tasksToLaunch release];
    }
    return allTasks;
}

- (void)setCommandEnvironment:(GLUPTaskRunner *)task withUserDefined:(BOOL)userDefined
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: setCommandEnvironment");
    if (userDefined) {
        [task setEnvironment:__environment];
    }
    [task setEnvironmentValue:[NSString stringWithFormat:@"%u", [ii2Prefs mainLogLevel]] forKey:@"II2LOGLEVEL"];
    [task setEnvironmentValue:[NSString stringWithFormat:@"%s", getlogin()] forKey:@"II2USER"];    
    [task setEnvironmentValue:[NSString stringWithString:([self needsAuthentication]?@"YES":@"NO")] forKey:@"II2AUTHENTICATED"];
    [task setEnvironmentValue:[self pkgTimestamp] forKey:@"II2PKGSTAMP"];
    [task setEnvironmentValue:[self pkgTimestampDisplay] forKey:@"II2PKGSTAMPDISPLAY"];
    [task setEnvironmentValue:gwBuildStamp() forKey:@"II2BUILDSTAMP"];
    [task setEnvironmentValue:gwBuildStampDisplay() forKey:@"II2BUILDSTAMPDISPLAY"];
    [task setEnvironmentValue:__dirName forKey:@"II2PKGDIR"];
    [task setEnvironmentValue:[self installLocation] forKey:@"II2INSTALLDIR"];
    [task setEnvironmentValue:[self urlString] forKey:@"II2PKGURL"];
    [task setEnvironmentValue:[[NSBundle mainBundle] resourcePath] forKey:@"II2RESOURCES"];
    if ([self isInteractive]) {
	[task setEnvironmentValue:[__cachedWindowController doServerName] forKey:@"II2DOSERVER"];
    }
    if ([self selectedSelectorSets]) {
	NSString *envValue = [[self selectedSelectorSets] componentsJoinedByString:@"|"];
	[task setEnvironmentValue:envValue forKey:@"II2SELECTEDSETS"];
    }
}

- (BOOL)checkForDownloadsInTaskArray:(NSArray *)tasks
{
    int i;

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: checkForDownloadsInTaskArray");
    if (tasks) {
	for (i=0; i<[tasks count]; i++) {
	    id <II2TaskDependencyProtocol> task = [tasks objectAtIndex:i];
	    if ([task isMemberOfClass:[II2DownloadTask class]]) {
		return YES;
	    }
	}
    }
    return NO;
}


// Workhorse. Assumes file available and md5-checked
- (NSArray *)__allocScriptRunTasksFor: (NSString *)scriptExtension
{
    NSMutableArray *tasks = [[NSMutableArray alloc] init];
    NSArray *scriptArgs = nil;
    GLUPTaskRunner *scriptTask = nil;
    NSString *scriptFileName = [[__baseName stringByAppendingString:@"."] stringByAppendingString:scriptExtension];
    NSNumber *scriptOutsize = [[__pkgPlist objectForKey:@"phaseoutputs"] objectForKey:scriptExtension];
    BOOL needsAuthentication = [self needsAuthentication] && ![scriptExtension isEqualToString:@"selector"];
    
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: __allocScriptRunTasksFor: %@", scriptExtension);
    scriptArgs = [NSArray arrayWithObjects: [II2TOOLFINDER pathForTool:@"/usr/bin/env"], [self fullPathForFileName:scriptFileName], nil];
    scriptTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:scriptArgs authenticator:(needsAuthentication?__auth:nil)];
    if (!scriptTask) {
	NSLog( @"Serious problem: could not create %@ run task.", scriptExtension);
	return nil;
    }
    [scriptTask setOutView:[__cachedWindowController outView]];
    [scriptTask setErrView:[__cachedWindowController errView]];
    [scriptTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: scriptExtension, @"type", scriptOutsize, @"expectedoutputsize", nil]];
    [self setCommandEnvironment:scriptTask withUserDefined:YES];
    [tasks addObject:scriptTask];
    if ([self isInteractive]) {
	II2ActionTask *startDOTask = [[II2ActionTask alloc] initWithTarget:__cachedWindowController owner:self selector:@"startDOServer" andObject:nil];
	II2ActionTask *stopDOTask = [[II2ActionTask alloc] initWithTarget:__cachedWindowController owner:self selector:@"stopDOServer" andObject:nil];
	[startDOTask addDependentTask:scriptTask];
	[scriptTask addDependentTask:stopDOTask];
	[tasks insertObject:startDOTask atIndex:0];
	[tasks addObject:stopDOTask];
    }
    {
	// We check them shortly before execution to minimize the window of opportunity for tampering.
	GLUPTaskRunner *md5Task = [self md5CheckTask:scriptFileName forTask:[tasks objectAtIndex:0]];
	if (!md5Task) {
	    NSLog( @"Serious problem: could not create %@ md5 task.", scriptExtension);
	    return nil;
	}
	[tasks insertObject:md5Task atIndex:0];
    }
    return tasks;
}

- (NSArray *)__allocPrepareRunTasks
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: __allocPrepareRunTasks");
    return [self __allocScriptRunTasksFor:@"prepare"];
}

- (NSArray *)__allocFixupRunTasks
{
    // Design decision: these should be linearly dependent first to last
    NSMutableArray *tasks = [[NSMutableArray alloc] init];
    NSArray *fixupArgs;
    GLUPTaskRunner *fixupTask;
    
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: __allocFixupRunTasks");
    fixupArgs = [NSArray arrayWithObjects: [II2TOOLFINDER pathForTool:@"/usr/bin/env"], @"/System/Library/CoreServices/FixupResourceForks", [self installLocation], nil];
    fixupTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:fixupArgs authenticator:([self needsAuthentication]?__auth:nil)];
    if (!fixupTask) {
	NSLog( @"Serious problem: could not create fixup run task.");
	return nil;
    }
    [fixupTask setOutView:[__cachedWindowController outView]];
    [fixupTask setErrView:[__cachedWindowController errView]];
    [fixupTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"fixup", @"type", nil]];
    [self setCommandEnvironment:fixupTask withUserDefined:NO];
    [tasks addObject:fixupTask];
    return tasks;
}

- (NSArray *)__allocUnarchiveRunTasks
{
    NSMutableArray *tasks = [[NSMutableArray alloc] init];
    NSString *unarchiver = [II2TOOLFINDER pathForTool:@"/usr/bin/gnutar"];;
    NSString *gzcat = [II2TOOLFINDER pathForTool:@"/usr/bin/gzcat"];
    NSString *bzcat = [II2TOOLFINDER pathForTool:@"/usr/bin/bzcat"];;
    NSArray *fixupTasks = nil;
    id lastTask = nil;
    NSString *fileNameRef = nil;
    II2ActionTask *unarchiveSentinelTask = [[II2ActionTask alloc] initWithTarget:self owner:self selector:@"unarchiveSentinel" andObject:nil];

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: __allocUnarchiveRunTasks");
    if ([self fileIsPartOfPackage:(fileNameRef = [__baseName stringByAppendingString:@".tar.bz2"])]) {
	NSMutableArray *fileSet = [[self allocFilePartsArrayFromFileName:fileNameRef filter:YES] autorelease];
        int i;

        for (i=0; i<[fileSet count]; i++) {
            NSArray *unarchiveArgs, *mkdirArgs = nil;
            GLUPTaskRunner *unarchiveTask, *mkdirTask = nil;
            NSString *partInstallLocation = [self installLocation];
            NSString *partName = [fileSet objectAtIndex:i];
	    NSString *partOutsize = [[__pkgPlist objectForKey:@"archiveoutputs"] objectForKey:partName];
            NSDictionary *subDirParts;
	    int partNumber = [self partNumberFromArchivePartFileName:partName];
            
	    if (partNumber == -1) {
		NSLog( @"Extremely weird program error: partNumber not found in file name %@", partName);
		[tasks release];
		return nil;
	    }
            if ((subDirParts = [__pkgPlist objectForKey:@"partsubdirs"])) {
                NSString *subdir = [subDirParts objectForKey:[NSString stringWithFormat:@"%d", partNumber]];
                
                if (subdir) {
		    // Subdirs that start with / are absolute dirs.
		    if ([subdir hasPrefix:@"/"]) {
			partInstallLocation = [NSString stringWithFormat:@"%@", subdir];
		    }
		    else {
			partInstallLocation = [NSString stringWithFormat:@"%@/%@", partInstallLocation, subdir];
		    }
		    mkdirArgs = [NSArray arrayWithObjects:@"/bin/mkdir", @"-p", partInstallLocation, nil];
		    mkdirTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:mkdirArgs authenticator:([self needsAuthentication]?__auth:nil)];
		    [mkdirTask setDirectoryPath:[self installLocation]];
		    [mkdirTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"ignore", @"type", nil]];
		    [mkdirTask setOutView:[__cachedWindowController outView]];
		    [mkdirTask setErrView:[__cachedWindowController errView]];
                }
            }
            
            unarchiveArgs = [NSArray arrayWithObjects:unarchiver, @"-C", partInstallLocation, @"--use-compress-program", bzcat, @"-xvpf", [self fullPathForFileName:partName], nil];
            unarchiveTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:unarchiveArgs authenticator:([self needsAuthentication]?__auth:nil)];
            [unarchiveTask setDirectoryPath:[self installLocation]];
            [unarchiveTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"unarchivepart", @"type", [fileSet objectAtIndex:i], @"archivepart", [NSNumber numberWithInt:partNumber], @"archivepartnumber", partOutsize, @"expectedoutputsize", @"NO", @"iserrdata", partInstallLocation, @"partinstalllocation", nil]];
            [unarchiveTask setOutView:[__cachedWindowController outView]];
            [unarchiveTask setErrView:[__cachedWindowController errView]];
            
	    if (mkdirTask) {
		if (lastTask) {
		    [lastTask addDependentTask:mkdirTask];
		}
		lastTask = mkdirTask;
	    }
	    {
		// We check them shortly before execution to minimize the window of opportunity for tampering.
		GLUPTaskRunner *md5Task = [self md5CheckTask:partName forTask:unarchiveTask];
		[tasks addObject:md5Task];
		if (lastTask) {
		    [lastTask addDependentTask:md5Task];
		}
		lastTask = md5Task;
	    }
	    if (mkdirTask) {
		[tasks addObject:mkdirTask];		
	    }
            [tasks addObject:unarchiveTask];
            if (lastTask) {
                [lastTask addDependentTask:unarchiveTask];
            }
            lastTask = unarchiveTask;
        }
    }
    if ([self fileIsPartOfPackage:(fileNameRef = [__baseName stringByAppendingString:@".tar.gz"])]) {
        NSMutableArray *fileSet = [[self allocFilePartsArrayFromFileName:fileNameRef filter:YES] autorelease];
        int i;
        
        for (i=0; i<[fileSet count]; i++) {
            NSArray *unarchiveArgs, *mkdirArgs = nil;
            GLUPTaskRunner *unarchiveTask, *mkdirTask = nil;
            NSString *partInstallLocation = [self installLocation];
            NSString *partName = [fileSet objectAtIndex:i];
	    NSString *partOutsize = [[__pkgPlist objectForKey:@"archiveoutputs"] objectForKey:partName];
            NSDictionary *subDirParts;
	    int partNumber = [self partNumberFromArchivePartFileName:partName];
            
            if ((subDirParts = [__pkgPlist objectForKey:@"partsubdirs"])) {
                NSString *subdir = [subDirParts objectForKey:[NSString stringWithFormat:@"%d", partNumber]];
                
                if (subdir) {
		    // Subdirs that start with / are absolute dirs.
		    if ([subdir hasPrefix:@"/"]) {
			partInstallLocation = [NSString stringWithFormat:@"%@", subdir];
		    }
		    else {
			partInstallLocation = [NSString stringWithFormat:@"%@/%@", partInstallLocation, subdir];
		    }
		    mkdirArgs = [NSArray arrayWithObjects:@"/bin/mkdir", @"-p", partInstallLocation, nil];
		    mkdirTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:mkdirArgs authenticator:([self needsAuthentication]?__auth:nil)];
		    [mkdirTask setDirectoryPath:[self installLocation]];
		    [mkdirTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"ignore", @"type", nil]];
		    [mkdirTask setOutView:[__cachedWindowController outView]];
		    [mkdirTask setErrView:[__cachedWindowController errView]];
                }
            }
            
            unarchiveArgs = [NSArray arrayWithObjects:unarchiver, @"-C", partInstallLocation, @"--use-compress-program", gzcat, @"-xvpf", [self fullPathForFileName:[fileSet objectAtIndex:i]], nil];
            unarchiveTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:unarchiveArgs authenticator:([self needsAuthentication]?__auth:nil)];
            [unarchiveTask setDirectoryPath:[self installLocation]];
            [unarchiveTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"unarchivepart", @"type", [fileSet objectAtIndex:i], @"archivepart", [NSNumber numberWithInt:partNumber], @"archivepartnumber", partOutsize, @"expectedoutputsize", @"NO", @"iserrdata", partInstallLocation, @"partinstalllocation", nil]];
            [unarchiveTask setOutView:[__cachedWindowController outView]];
            [unarchiveTask setErrView:[__cachedWindowController errView]];

	    if (mkdirTask) {
		if (lastTask) {
		    [lastTask addDependentTask:mkdirTask];
		}
		lastTask = mkdirTask;
	    }
	    
	    {
		// When our ToC has been tested against tampering, we can trust the MD5's and we check them shortly before execution to minimize the window of opportunity for tampering.
		GLUPTaskRunner *md5Task = [self md5CheckTask:partName forTask:unarchiveTask];
		[tasks addObject:md5Task];
		if (lastTask) {
		    [lastTask addDependentTask:md5Task];
		}
		lastTask = md5Task;
	    }
            [tasks addObject:unarchiveTask];
            if (lastTask) {
                [lastTask addDependentTask:unarchiveTask];
            }
            lastTask = unarchiveTask;
        }
    }

    if ([self needsResourceForksFixup]) {
	fixupTasks = [self __allocFixupRunTasks];
	if (!fixupTasks) {
	    return nil; // Warning already performed
	}
	// Last (fixup) task takes place of unarchive tasks (for didFinish handling)
	lastTask = [fixupTasks lastObject];
    }
    if (fixupTasks) {
	if ([fixupTasks count]) {
	    [tasks makeObjectsPerformSelector:@selector(addDependentTask:) withObject:[fixupTasks objectAtIndex:0]];
	    [tasks addObjectsFromArray:fixupTasks];
	}
	[fixupTasks release];
    }
    // All are ignore tasks, except for last task which signals all unarchiving and fixup tasks are done
    [lastTask addDependentTask:unarchiveSentinelTask];
    [tasks addObject:unarchiveSentinelTask];

    return tasks;
}


// Workhorse. Assumes file available and md5-checked
- (NSArray *)__allocConfigureRunTasks
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: __allocConfigureRunTasks");
    return [self __allocScriptRunTasksFor:@"configure"];
}

- (NSArray *)__allocInstallRunTasks
{
    NSMutableArray *allTasks = [[NSMutableArray alloc] init];
    NSArray *prepareRunTasks = nil;
    NSArray *unarchiveRunTasks = nil;
    NSArray *configureRunTasks = nil;

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: __allocInstallRunTasks");
    if ([self fileIsPartOfPackage:[__baseName stringByAppendingString:@".configure"]]) {
	configureRunTasks = [self __allocConfigureRunTasks];
	if (!configureRunTasks || ([configureRunTasks count] == 0)) {
	    NSLog( @"Serious problem: could not create configuration tasks.");
	    return nil;
	}
    }
    if ([self fileIsPartOfPackage:[__baseName stringByAppendingString:@".tar.gz"]] || [self fileIsPartOfPackage:[__baseName stringByAppendingString:@".tar.bz2"]]) {
	unarchiveRunTasks = [self __allocUnarchiveRunTasks];
	if (!unarchiveRunTasks || ([unarchiveRunTasks count] == 0)) {
	    NSLog( @"Serious problem: could not create unarchive tasks.");
	    return nil;
	}
	if (configureRunTasks) {
	    [unarchiveRunTasks makeObjectsPerformSelector:@selector(addDependentTask:) withObject:[configureRunTasks objectAtIndex:0]];
	}
    }
    if ([self fileIsPartOfPackage:[__baseName stringByAppendingString:@".prepare"]]) {
	prepareRunTasks = [self __allocPrepareRunTasks];
	if (!prepareRunTasks || ([prepareRunTasks count] == 0)) {
	    NSLog( @"Serious problem: could not create preparation tasks.");
	    return nil;
	}
	if (unarchiveRunTasks) {
	    [[prepareRunTasks lastObject] addDependentTasks:unarchiveRunTasks];
	}
	else if (configureRunTasks) {
	    [[prepareRunTasks lastObject] addDependentTask:[configureRunTasks objectAtIndex:0]];
	}
    }
    // Combine the arrays (keep order?)
    if (prepareRunTasks) {
	if ([prepareRunTasks count]) {
	    [allTasks addObjectsFromArray:prepareRunTasks];
	}
	[prepareRunTasks release];
    }
    if (unarchiveRunTasks) {
	if ([unarchiveRunTasks count]) {
	    [allTasks addObjectsFromArray:unarchiveRunTasks];
	}
	[unarchiveRunTasks release];
    }
    if (configureRunTasks) {
	if ([configureRunTasks count]) {
	    [allTasks addObjectsFromArray:configureRunTasks];
	}
	[configureRunTasks release];
    }
    return allTasks;
}

- (NSArray *)__allocRemoveRunTasks
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: __allocRemoveRunTasks");
    return [self __allocScriptRunTasksFor:@"remove"];
}

- (NSArray *)__allocSelectorRunTasks
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: __allocSelectorRunTasks");
    return [self __allocScriptRunTasksFor:@"selector"];
}

- (BOOL)prepareInstallDir
{
    BOOL isDir = NO;

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: prepareInstallDir");
    if ([self needsAuthentication]) {
	if (![__auth setPrivileged:YES]) {
	    // do nothing, authentication failed
	    NSLog( @"Authentication failed on prepare directory");
	    return NO;
	}
    }
    if (![DEFFM fileExistsAtPath:[self installLocation] isDirectory:&isDir]) {
	NSArray *mkdirArgs = [NSArray arrayWithObjects: @"/bin/mkdir", @"-p", [self installLocation], nil];
	GLUPTaskRunner *mkdirTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:mkdirArgs authenticator:([self needsAuthentication]?__auth:nil)];
	if (!mkdirTask) {
	    return NO;
	}
	[mkdirTask run];
	if (![DEFFM fileExistsAtPath:[self installLocation] isDirectory:&isDir]) {
	    NSLog( @"Failed to create %s", [[self installLocation] fileSystemRepresentation]);
	    return NO;
	}
    }
    if (!isDir) {
	NSLog( @"%s is not a directory", [[self installLocation] fileSystemRepresentation]);
    }
    return isDir;
}

// This one does all: selector, install, configure and remove
- (NSArray *)allocTasksForAction:(NSString *)action
{
    NSMutableArray *filesToCheck = [[[NSMutableArray alloc] init] autorelease];
    NSMutableArray *allTasks =  [[NSMutableArray alloc] init];
    NSArray *commandTasks = nil;
    int i;
    id <II2TaskDependencyProtocol> firstRunTask = nil;
    NSString *fileNameRef;

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: allocTasksForAction %@", action);
    if ([action isEqualToString:@"trueinstall"]) {
	NSArray *configureDependencies = [self dependentFilesForStage:@"configure" filter:YES];
	NSArray *prepareDependencies = [self dependentFilesForStage:@"prepare" filter:YES];
	commandTasks = [self __allocInstallRunTasks];
	if ([self fileIsPartOfPackage:[__baseName stringByAppendingString:@".prepare"]]) {
	    [filesToCheck addObject:[__baseName stringByAppendingString:@".prepare"]];
	    if (prepareDependencies && [prepareDependencies count]) {
		[filesToCheck addObjectsFromArray:prepareDependencies];
	    }
	}
	if ([self fileIsPartOfPackage:(fileNameRef = [__baseName stringByAppendingString:@".tar.bz2"])]) {
	    NSMutableArray *partSet = [self allocFilePartsArrayFromFileName:fileNameRef filter:YES];
	    if (partSet) {
		if ([partSet count]) {
		    [filesToCheck addObjectsFromArray:partSet];
		}
		[partSet release];
	    }
	}
	if ([self fileIsPartOfPackage:(fileNameRef = [__baseName stringByAppendingString:@".tar.gz"])]) {
	    NSMutableArray *partSet = [self allocFilePartsArrayFromFileName:fileNameRef filter:YES];
	    if (partSet) {
		if ([partSet count]) {
		    [filesToCheck addObjectsFromArray:partSet];
		}
		[partSet release];
	    }
	}
	if ([self fileIsPartOfPackage:[__baseName stringByAppendingString:@".configure"]]) {
	    int i;
	    [filesToCheck addObject:[__baseName stringByAppendingString:@".configure"]];
	    if (configureDependencies && [configureDependencies count]) {
		[filesToCheck addObjectsFromArray:configureDependencies];
	    }
	    // Find the configure script task itself and set an environment variable for it telling the script we are running configure separately
	    for (i=0; i<[commandTasks count]; i++) {
		id <II2TaskPropertiesProtocol, NSObject> task = [commandTasks objectAtIndex:i];
		if ([task isKindOfClass:[GLUPTaskRunner class]]) {
		    NSString *type = [task propertyForKey:@"type"];
		    if ([type isEqualToString:@"configure"]) {
			[(GLUPTaskRunner *)task setEnvironmentValue:@"install" forKey:@"II2USERACTION"];
		    }
		}
	    }
	}
    }
    else if ([action isEqualToString:@"configure"]) {
	int i;
	NSArray *configureDependencies = [self dependentFilesForStage:@"configure" filter:YES];
	commandTasks = [self __allocConfigureRunTasks];
	if ([self fileIsPartOfPackage:[__baseName stringByAppendingString:@".configure"]]) {
	    [filesToCheck addObject:[__baseName stringByAppendingString:@".configure"]];
	    if (configureDependencies && [configureDependencies count]) {
		[filesToCheck addObjectsFromArray:configureDependencies];
	    }
	}
	// Find the configure script task itself and set an environment variable for it telling the script we are running configure separately
	for (i=0; i<[commandTasks count]; i++) {
	    id <II2TaskPropertiesProtocol, NSObject> task = [commandTasks objectAtIndex:i];
	    if ([task isKindOfClass:[GLUPTaskRunner class]]) {
		NSString *type = [task propertyForKey:@"type"];
		if ([type isEqualToString:@"configure"]) {
		    [(GLUPTaskRunner *)task setEnvironmentValue:@"configure" forKey:@"II2USERACTION"];
		}
	    }
	}
    }
    else if ([action isEqualToString:@"remove"]) {
	NSArray *removeDependencies = [self dependentFilesForStage:@"remove" filter:YES];
	commandTasks = [self __allocRemoveRunTasks];
	if ([self fileIsPartOfPackage:[__baseName stringByAppendingString:@".remove"]]) {
	    [filesToCheck addObject:[__baseName stringByAppendingString:@".remove"]];
	    if (removeDependencies && [removeDependencies count]) {
		[filesToCheck addObjectsFromArray:removeDependencies];
	    }
	}
    }
    else if ([action isEqualToString:@"selectandinstall"]) {
	NSArray *selectorDependencies = [self dependentFilesForStage:@"selector" filter:YES];
	commandTasks = [self __allocSelectorRunTasks];
	if ([self fileIsPartOfPackage:[__baseName stringByAppendingString:@".selector"]]) {
	    [filesToCheck addObject:[__baseName stringByAppendingString:@".selector"]];
	    if (selectorDependencies && [selectorDependencies count]) {
		[filesToCheck addObjectsFromArray:selectorDependencies];
	    }
	}
    }
    else if ([action isEqualToString:@"fatten"]) {
	commandTasks = [self __allocFattenTasks];
	// filesToCheck remains empty
	// commandTasks is empty when everything has been checked already
    }
    if (commandTasks && [commandTasks count]) {
	firstRunTask = [commandTasks objectAtIndex:0];
    }
    for (i=0; i<[filesToCheck count]; i++) {
	NSString *fileName = [filesToCheck objectAtIndex:i];
	NSArray *tasksToLaunch = [self allocTasksForFileAvailable:fileName forTask:firstRunTask];
	if (!tasksToLaunch) {
	    [self destroyTaskArray:commandTasks];
	    [self destroyTaskArray:allTasks];
	    return nil;
	}
	if ([tasksToLaunch count]) {
	    [allTasks addObjectsFromArray:tasksToLaunch];
	}
	[tasksToLaunch release];
    }
    
    if (commandTasks) {
	if ([commandTasks count]) {
	    [allTasks addObjectsFromArray:commandTasks];
	}
	[commandTasks release];
    }
    return allTasks;
}

- (void)destroyTaskArray:(NSArray *)tasks
{
    int i;

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: destroyTaskArray");
    for (i=0; i<[tasks count]; i++) {
	id task = [tasks objectAtIndex:i];
	[task release];
    }
    [tasks release];
}

- (void)markTaskArray:(NSArray *)tasks forAction:(NSString *)action
{
    int i;

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: markTaskArray");
    for (i=0; i<[tasks count]; i++) { 
	id task = [tasks objectAtIndex:i];
	if ([task conformsToProtocol:@protocol( II2TaskPropertiesProtocol)]) {
	    [(id <II2TaskPropertiesProtocol>)task setProperty:action forKey:@"useraction"];
	}
    }
}

- (void)doActionTasks:(NSArray *)tasks
{
    int i;
    // Chain all download tasks so only one is executed at the time
    id <II2TaskDependencyProtocol> lastTask = nil;
    unsigned int totalDownloadTasksOutputSize = 0;
    unsigned int totalActionTasksOutputSize = 0;

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: doActionTasks %@", [tasks description]);
    for (i=0; i<[tasks count]; i++) {
	id <II2TaskDependencyProtocol> task = [tasks objectAtIndex:i];
	if ([task isKindOfClass:[II2DownloadTask class]]) {
	    II2TimestampComparisonResult compare;
	    [self readRemotePkgPlist:YES]; // Force read, there is still a race condition here
	    compare = [self remoteStatus];
	    if (compare != NSOrderedSame) {
		// Remote is different version and we have downloads
		NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];
		
		NSLog( @"It is not possible to execute the current set of actions. Not all necessary parts are available. Downloading these would man downloading parts from the newer package version into your current version and this is a forbidden combination.\n\nThis happens for instance when you choose Install and though the selector phase is available, the files necessary for the actual install are not.");
		
		[sheetDict setObject:self forKey:@"resulthandler"];
		[sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
		[sheetDict setObject:@"Alert" forKey:@"type"];
		[sheetDict setObject:@"Warning" forKey:@"subtype"];
		[sheetDict setObject:@"Missing files" forKey:@"title"];
		[sheetDict setObject:@"It is not possible to execute the current set of actions. Not all necessary parts are available. Downloading these would man downloading parts from the newer package version into your current version and this is a forbidden combination.\n\nThis happens for instance when you choose Install and though the selector phase is available, the files necessary for the actual install are not." forKey:@"info"];
		[sheetDict setObject:@"Accept" forKey:@"okbutton"];
		
		[[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
		[__cachedWindowController autoButtonsEnabled];
		[self destroyTaskArray:tasks];
		return;
	    }
	    if (![ii2Prefs fastNetwork] && lastTask) {
		// Only chain them when fast network is turned off.
		[lastTask addDependentTask:task];
	    }
	    lastTask = task;
	}
    }

    // First, get total output size for reporting progress
    for (i=0; i<[tasks count]; i++) {
	id <II2TaskSetMemberProtocol, II2TaskStatusProviderProtocol, II2TaskPropertiesProtocol> task = [tasks objectAtIndex:i];
	if ([task isKindOfClass:[II2DownloadTask class]])  {
	    totalDownloadTasksOutputSize += [task sizeOfTaskWhenCompleted];
	}
	else if ([task isKindOfClass:[GLUPTaskRunner class]] && [[task properties] objectForKey:@"expectedoutputsize"]) {
	    totalActionTasksOutputSize += [task sizeOfTaskWhenCompleted];
	}
    }
    [__cachedWindowController setReportMaxDownloadsSize:totalDownloadTasksOutputSize];
    [__cachedWindowController setReportMaxActionsSize:totalActionTasksOutputSize];

    // Then, put the tasks into the set and set them up for reporting
    for (i=0; i<[tasks count]; i++) {
	id <II2TaskSetMemberProtocol, II2TaskStatusProviderProtocol, II2TaskPropertiesProtocol> task = [tasks objectAtIndex:i];
	// This retains them:
	[__ii2RunningTaskSet addII2Task:task];
	if ([task isKindOfClass:[II2DownloadTask class]]) {
	    // Set up for progress reporting
	    [task setReportTarget:__cachedWindowController];
	    [task setReportIncrementTreshold:totalDownloadTasksOutputSize/4096];
	}
	else if ([task isKindOfClass:[GLUPTaskRunner class]] && [[task properties] objectForKey:@"expectedoutputsize"]) {
	    // Set up for progress reporting
	    [task setReportTarget:__cachedWindowController];
	    [task setReportIncrementTreshold:totalActionTasksOutputSize/64];
	}
    }
    
    [__cachedWindowController startDisplayDownloadProgress:totalDownloadTasksOutputSize];
    [__cachedWindowController startDisplayActionProgress:totalActionTasksOutputSize];

    // Then launch
    for (i=[tasks count]-1; i>=0; i--) {
	id <II2TaskSetMemberProtocol> task = [tasks objectAtIndex:i];
	// Reverse order to prevent race condition where a task endWaits a dependent task after which the dependent task also gets a launch from us, effectively launching it twice.
	[task launch];
    }
}

// This one does only install, configure and remove, it is also called when the selector task finishes
- (void)action:(NSString *)action
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: action: %@", action);

    if ([action isEqualToString:@"selectandinstall"]) {
	if ([self hasSelectorScript]) {
	    NSArray *allTasks = [self allocTasksForAction:action];
	    if (allTasks) {
		if ([self checkForDownloadsInTaskArray:allTasks]  && ![self canBeFattened]) {
		    NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];
		    
		    NSLog( @"It is not possible to download missing parts of packages when network access has been disabled in your preferences, or the package is locked or not writable. Check your preferences or lock status or copy the package to a writable location first.");

		    [sheetDict setObject:self forKey:@"resulthandler"];
		    [sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
		    [sheetDict setObject:@"Alert" forKey:@"type"];
		    [sheetDict setObject:@"Warning" forKey:@"subtype"];
		    [sheetDict setObject:@"Cannot Fatten" forKey:@"title"];
		    [sheetDict setObject:@"It is not possible to download missing parts of packages when network access has been disabled in your preferences, or the package is locked or not writable. Check your preferences or lock status or copy the package to a writable location first." forKey:@"info"];
		    [sheetDict setObject:@"Accept" forKey:@"okbutton"];
		    
		    [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
		    [__cachedWindowController autoButtonsEnabled];
		    [self destroyTaskArray:allTasks];
		    return;
		}
		[[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Starting %@ tasks\n", action] attributes:[self infoAttributes]] autorelease]];
		[self markTaskArray:allTasks forAction:action];
		[self doActionTasks:allTasks];
		[allTasks release];
	    }		
	    else {
		[__cachedWindowController autoButtonsEnabled];
	    }
	}
	else {
	    [self action:@"dependencycheckandinstall"];
	}
	return;
    }
    else if ([action isEqualToString:@"dependencycheckandinstall"] && [self hasArchive]) {
	if ([self reportPkgDependencies]) {
	    [self action:@"trueinstall"];
	}
	else {
	    NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];
	    
	    [sheetDict setObject:self forKey:@"resulthandler"];
	    [sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
	    [sheetDict setObject:@"Alert" forKey:@"type"];
	    [sheetDict setObject:@"RequiredOthers" forKey:@"subtype"];
	    [sheetDict setObject:@"dependencycheckandinstall" forKey:@"action"];
	    [sheetDict setObject:@"Install/deinstall dependencies first" forKey:@"title"];
	    [sheetDict setObject:@"There are dependency problems. Either required files are unavailable or forbidden files are available. You should act (if possible, see the Readme view) on the reported dependency problems (see Message View) before you install this package. Only if you really know what you are doing (e.g. you are a unix-savvy geek with a special setup), press Ignore to install anyway." forKey:@"info"];
	    [sheetDict setObject:@"Cancel" forKey:@"okbutton"];
	    [sheetDict setObject:@"Ignore" forKey:@"otherbutton"];
	    
	    [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
	}
    }
    else if (([action isEqualToString:@"trueinstall"] && [self hasArchive]) || ([action isEqualToString:@"remove"] && [self hasRemoveScript]) || ([action isEqualToString:@"configure"] && [self hasConfigureScript]))
    {
	NSArray *allTasks = [self allocTasksForAction:action];
	if (allTasks) {
	    if ([self checkForDownloadsInTaskArray:allTasks]  && ![self canBeFattened]) {
		NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];
		
		NSLog( @"It is not possible to download missing parts of packages when network access has been disabled in your preferences, or the package is locked or not writable. Check your preferences or lock status or copy the package to a writable location first.");
		
		[sheetDict setObject:self forKey:@"resulthandler"];
		[sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
		[sheetDict setObject:@"Alert" forKey:@"type"];
		[sheetDict setObject:@"Warning" forKey:@"subtype"];
		[sheetDict setObject:@"Cannot Fatten" forKey:@"title"];
		[sheetDict setObject:@"It is not possible to download missing parts of packages when network access has been disabled in your preferences, or the package is locked or not writable. Check your preferences or lock status or copy the package to a writable location first." forKey:@"info"];
		[sheetDict setObject:@"Accept" forKey:@"okbutton"];
		[[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
		[__cachedWindowController autoButtonsEnabled];
		[self destroyTaskArray:allTasks];
		return;
	    }
	    if ([self needsAuthentication]) {
		if (![__auth setPrivileged:YES]) {
		    // do nothing, authentication failed
		    if ([ii2Prefs mainLogLevel] > 0) NSLog( @"Authentication failed on %@", action);
		    [__cachedWindowController autoButtonsEnabled];
		    [self destroyTaskArray:allTasks];
		    return;
		}
	    }
	    if ([action isEqualToString:@"trueinstall"] && ![self prepareInstallDir]) {
		[__cachedWindowController autoButtonsEnabled];
		[self destroyTaskArray:allTasks];
		return;
	    }
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Starting %@ tasks\n", action] attributes:[self infoAttributes]] autorelease]];
	    [self markTaskArray:allTasks forAction:action];
	    [self doActionTasks:allTasks];
	    [allTasks release];
	}
	else {
	    [__cachedWindowController autoButtonsEnabled];
	}
    }
    else if ([action isEqualToString:@"fatten"]) {
	[self fatten];
    }
}

- (BOOL)update
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: update");
    if ([self canBeFattened]) {
	if (![self readRemotePkgPlist:YES]) {
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"The retrieval of the remote ToC failed.\n"] attributes:[self signalAttributes]] autorelease]];
	    return NO;
	}
	if (![self remoteUpdateTimestamp]) {
	    NSString *plistName = [__baseName stringByAppendingString:@".plist"];
	    NSString *newPlistName = [__baseName stringByAppendingString:@".plist.new"];
	    NSString *plistMD5Name = [__baseName stringByAppendingString:@".plist.md5"];
	    NSString *remoteNewMD5URLString = [self fullURLPathForFileName:plistMD5Name];
	    NSString *newPlistContents = [self foregroundGetURLString:plistName];
	    if (newPlistContents && __remotePkgPlist) {
		if ([[__remotePkgPlist objectForKey:@"timestamp"] isEqualToString:[self pkgTimestamp]]) {
		    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"The remote package has the same timestamp as the local package. No update required.\n"] attributes:[self infoAttributes]] autorelease]];
		    [__cachedWindowController autoButtonsEnabled];
		    return YES;
		}
		if ([newPlistContents writeToFile:[self fullPathForFileName:newPlistName] atomically:YES]) {
		    if (!checkMD5OfFileAgainstContentsOfURL( [self fullPathForFileName:newPlistName], remoteNewMD5URLString)) {
			[[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"The check of the remote package md5 checksum failed.\n"] attributes:[self signalAttributes]] autorelease]];
		    }
		    else {
			NSString *sigFile = [self fullPathForFileName:[__baseName stringByAppendingString:@".plist.sig"]];
			NSString *sigURLString = [self fullURLPathForFileName:[__baseName stringByAppendingString:@".plist.sig"]];
			NSURL *sigURL = [NSURL URLWithString:sigURLString];
			NSData *sigData = urlDataFromURL( sigURL);
			NSMutableArray *toBeRemoved = [[[NSMutableArray alloc] init] autorelease];
			NSDictionary *oldFilesDict = [__pkgPlist objectForKey:@"files"];
			NSDictionary *newFilesDict = [__remotePkgPlist objectForKey:@"files"];
			NSEnumerator *enumerator = [oldFilesDict keyEnumerator];
			NSString *oldFileKey;

			while ((oldFileKey = [enumerator nextObject])) {
			    NSDictionary *newFileDict = [newFilesDict objectForKey:oldFileKey];
			    if ([oldFileKey isEqualToString:@"Custom.plist"]) {
				continue;
			    }
			    if (newFileDict) {
				// File in both old and new
				NSDictionary *oldFileDict = [oldFilesDict objectForKey:oldFileKey];
				NSString *oldmd5 = [oldFileDict objectForKey:@"md5"];
				NSString *newmd5 = [newFileDict objectForKey:@"md5"];
				if (!newmd5) {
				    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"The new Md5 is missing for file %@. Please retry later.\n", oldFileKey] attributes:[self signalAttributes]] autorelease]];
				    return NO;
				}
				if (![oldmd5 isEqualToString:newmd5]) {
				    if ([self fileIsAvailable:oldFileKey]) {
					[toBeRemoved addObject:oldFileKey];
					[[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"File %@ has changed and will be removed from the local i-Package. It will be downloaded again when required.\n", oldFileKey] attributes:[self infoAttributes]] autorelease]];
					if ([ii2Prefs mainLogLevel] > 6) NSLog( @"Tagging %s for removal", [[self fullPathForFileName:oldFileKey] fileSystemRepresentation]);
				    }
				}
			    }
			    else {
				// File only in old
				if ([self fileIsAvailable:oldFileKey]) {
				    [toBeRemoved addObject:oldFileKey];
				    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"File %@ is not part of the updated i-Package and will be removed from the local i-Package.\n", oldFileKey] attributes:[self infoAttributes]] autorelease]];
				    if ([ii2Prefs mainLogLevel] > 6) NSLog( @"Tagging %s for removal", [[self fullPathForFileName:oldFileKey] fileSystemRepresentation]);
				}
			    }
			}
			// When we get here, we know what to do: write new plist into old and replace __pkgPlist, remove certain files and possibly download others.
			if ([toBeRemoved count]) {
			    int i;
			    for (i=0; i<[toBeRemoved count]; i++) {
				[self deleteFile:[toBeRemoved objectAtIndex:i]];
			    }
			}
			[newPlistContents writeToFile:[self fullPathForFileName:plistName] atomically:YES];
			[DEFFM removeFileAtPath:sigFile handler:nil];
			if (sigData) {
			    if (![sigData writeToFile:sigFile atomically:NO]) {
				if ([ii2Prefs mainLogLevel] > 0) NSLog( @"Unable to write sig file %@. Sig %@ probably unavailable.", sigFile, sigURLString);
			    }
			}
			[DEFFM removeFileAtPath:newPlistName handler:nil];
			[__pkgPlist release];
			__pkgPlist = [__remotePkgPlist retain];
			[self setupFileTableSourceAndInfo];
			[__cachedWindowController displayPackage];
			[self showInfo];
			return YES;
		    }
		}
		else {
		    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Cannot write the new plist.\n"] attributes:[self signalAttributes]] autorelease]];
		}
	    }
	    else {
		[[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Plist is missing from the remote i-Package. Please retry later.\n"] attributes:[self signalAttributes]] autorelease]];
	    }
	}
	else {
	    NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"The remote i-Package is currently being updated. Please retry later.\n"] attributes:[self signalAttributes]] autorelease]];
	    
	    [sheetDict setObject:self forKey:@"resulthandler"];
	    [sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
	    [sheetDict setObject:@"Alert" forKey:@"type"];
	    [sheetDict setObject:@"Warning" forKey:@"subtype"];
	    [sheetDict setObject:@"Temporarily unavailable" forKey:@"title"];
	    [sheetDict setObject:@"The remote i-Package is being updated. Please retry later." forKey:@"info"];
	    [sheetDict setObject:@"Accept" forKey:@"okbutton"];
	    [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
	}
    }
    else {
	NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];
	[[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"It is not possible to update packages when network access has been disabled in your preferences or the package is locked or not writable.\n"] attributes:[self signalAttributes]] autorelease]];
	
	[sheetDict setObject:self forKey:@"resulthandler"];
	[sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
	[sheetDict setObject:@"Alert" forKey:@"type"];
	[sheetDict setObject:@"Warning" forKey:@"subtype"];
	[sheetDict setObject:@"Disabled" forKey:@"title"];
	[sheetDict setObject:@"It is not possible to update packages when network access has been disabled in your preferences or the package is locked or not writable." forKey:@"info"];
	[sheetDict setObject:@"Accept" forKey:@"okbutton"];
	[[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
    }
    [__cachedWindowController autoButtonsEnabled];
    return NO;
}

- (void)__inspect:(NSString *)fileName inView:(NSTextView *)aView
{
    NSString *unarchiver = [II2TOOLFINDER pathForTool:@"/usr/bin/gnutar"];
    NSString *ext = [fileName pathExtension];

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: Inspecting: %s", [fileName fileSystemRepresentation]);
    [aView resetParagraphStyleAndClear];

    // The real thing, file is available and has valid md5 and type info
    if ([fileName hasSuffix:@".gz"] || [fileName hasSuffix:@".bz2"]) {
	NSString *decompressor = [ext isEqualToString:@"gz"] ? [II2TOOLFINDER pathForTool:@"/usr/bin/gzcat"] : [II2TOOLFINDER pathForTool:@"/usr/bin/bzcat"];

	if ([[self fileTypeDisplayString:fileName] hasSuffix:@"tar archive"]) {
	    // Possible parts
	    NSMutableArray *partSet = [self allocFilePartsArrayFromFileName:fileName filter:NO];
	    NSMutableArray *tasksToRun = [[NSMutableArray alloc] init];
	    int i;
	    NSDictionary *subDirParts = [__pkgPlist objectForKey:@"partsubdirs"];

	    for (i=0; i<[partSet count]; i++) {
		NSArray *setDecompressArgs = [NSArray arrayWithObjects: decompressor, [partSet objectAtIndex:i], nil];
		GLUPTaskRunner *setDecompressTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:setDecompressArgs authenticator:nil];
		NSArray *setUnarchiveArgs = [NSArray arrayWithObjects: unarchiver, @"tvf", @"-", nil];
		GLUPTaskRunner *setUnarchiveTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:setUnarchiveArgs authenticator:nil];
		id <II2TaskDependencyProtocol> lastTask = [tasksToRun lastObject];
		NSString *subDir = [subDirParts objectForKey:[NSString stringWithFormat:@"%d", [self partNumberFromArchivePartFileName:[partSet objectAtIndex:i]]]];
		
		if (!subDir) {
		    subDir = @"[Problem: subdir not found]";
		}
		if (![subDir hasPrefix:@"/"]) {
		    subDir = [NSString stringWithFormat:@"%@/%@", [self installLocation], subDir];
		}

		[setDecompressTask setDirectoryPath:__dirName];
		[setDecompressTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"ignore", @"type", nil]];
		[setUnarchiveTask setOutView:aView];
		[setUnarchiveTask setErrView:[__cachedWindowController errView]];
		[setDecompressTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"ignore", @"type", [NSString stringWithFormat:@"The following part (%@) will be unarchived in directory %@\n", [NSString stringWithFormat: @"part %d", i], subDir], @"stdoutmessage", nil]];
		[setUnarchiveTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"textinspect", @"type", fileName, @"filename", nil]];
		[setDecompressTask connectOutputTo:setUnarchiveTask];
		[tasksToRun addObject:setDecompressTask];
		[tasksToRun addObject:setUnarchiveTask];
		if (lastTask) {
		    [lastTask addDependentTask:setDecompressTask];
		}
	    }
	    [partSet release];
	    [self doActionTasks:tasksToRun];
	    [tasksToRun release];
	}
	else {
	    NSArray *decompressArgs = [NSArray arrayWithObjects: decompressor, fileName, nil];
	    GLUPTaskRunner *decompressTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:decompressArgs authenticator:nil];

	    [decompressTask setDirectoryPath:__dirName];

	    if ([[self fileTypeDisplayString:fileName] isEqualToString:@"Rich Text Format data, version 1, Apple Macintosh"]) {
		NSMutableData *outData = [[[NSMutableData alloc] init] autorelease];

		[aView setRichText:YES];
		[aView setUsesRuler:YES];
		[decompressTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"rtfinspect", @"type", outData, @"outdata", fileName, @"filename", aView, @"view", nil]];
		[decompressTask setOutData:outData];
		[__ii2RunningTaskSet addII2Task:decompressTask];
		[decompressTask launch];
	    }
	    else if ([[self fileTypeDisplayString:fileName] hasSuffix:@"text"]) {
		[decompressTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"ignore", @"type", nil]];
		[decompressTask setOutView:aView];
		[decompressTask setErrView:[__cachedWindowController errView]];
		[__ii2RunningTaskSet addII2Task:decompressTask];
		[decompressTask launch];
	    }
	    else {
		// Two tasks in a pipe, uncompress and hexdump
		NSArray *hexdumpArgs = [NSArray arrayWithObjects: [II2TOOLFINDER pathForTool:@"/usr/bin/hexdump"], @"-C", nil];
		GLUPTaskRunner *hexdumpTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:hexdumpArgs authenticator:nil];

		[hexdumpTask setOutView:aView];
		[hexdumpTask setErrView:[__cachedWindowController errView]];
		[decompressTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"ignore", @"type", nil]];
		[hexdumpTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"ignore", @"type", nil]];
		[decompressTask connectOutputTo:hexdumpTask];
		[__ii2RunningTaskSet addII2Task:hexdumpTask];
		[hexdumpTask launch];
		[__ii2RunningTaskSet addII2Task:decompressTask];
		[decompressTask launch];
	    }
	}
    }
    else if ([[self fileTypeDisplayString:fileName] isEqualToString:@"Rich Text Format data, version 1, Apple Macintosh"]) {
	[aView setRichText:YES];
	[aView setUsesRuler:YES];
	[aView selectAll:nil];
	[aView replaceCharactersInRange:[aView selectedRange] withRTF:[self dataFromFileContents:fileName]];
	[__cachedWindowController autoButtonsEnabled];
    }
    else if ([[self fileTypeDisplayString:fileName] hasSuffix:@"text"]) {
	// This is a complex workaround because a setString still obeys previous ruler settings
	[[aView textStorage] appendAttributedString: [[[NSAttributedString alloc]
                             initWithString: [self stringFromFileContents:fileName] attributes:[aView fixedNormalAttributes]] autorelease]];

	[__cachedWindowController autoButtonsEnabled];
    }
    else if ([[self fileTypeDisplayString:fileName] hasPrefix:@"Zip archive data"]) {
	NSArray *unzipArgs = [NSArray arrayWithObjects: [II2TOOLFINDER pathForTool:@"/usr/bin/unzip"], @"-l", fileName, nil];
	GLUPTaskRunner *unzipTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:unzipArgs authenticator:nil];

	[unzipTask setDirectoryPath:__dirName];
	[unzipTask setOutView:aView];
	[unzipTask setErrView:[__cachedWindowController errView]];
	[unzipTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"ignore", @"type", nil]];
	[__ii2RunningTaskSet addII2Task:unzipTask];
	[unzipTask launch];
    }
    else {
	NSArray *hexdumpArgs = [NSArray arrayWithObjects: [II2TOOLFINDER pathForTool:@"/usr/bin/hexdump"], @"-C", fileName, nil];
	GLUPTaskRunner *hexdumpTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:hexdumpArgs authenticator:nil];

	[hexdumpTask setDirectoryPath:__dirName];
	[hexdumpTask setOutView:aView];
	[hexdumpTask setErrView:[__cachedWindowController errView]];
	[hexdumpTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"ignore", @"type", nil]];
	[__ii2RunningTaskSet addII2Task:hexdumpTask];
	[hexdumpTask launch];
    }
    if (aView == [__cachedWindowController readmeView]) {
	[__cachedWindowController showTab:ii2ReadmeTabTag];
    }
    else if (aView == [__cachedWindowController inspectView]) {
	[__cachedWindowController showTab:ii2ContentsTabTag];
    }
}

- (void)inspect:(NSString *)fileName withModifiers:(unsigned int)modifiers inView:(NSTextView *)aView
{
    NSString *useFileName = [self fileNameForInspect:fileName withModifiers:modifiers];
    II2ActionTask *actionTask;

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: Running inspect on %s", [fileName fileSystemRepresentation]);

    actionTask = [[II2ActionTask alloc] initWithTarget:self owner:self selector:@"__inspect:inView:" andObject:useFileName andObject:aView];

    if (![self fileSetIsAvailable:useFileName]) {
	if (![self canBeFattened]) {
	    NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];
	    if ([ii2Prefs mainLogLevel] > 0) NSLog( @"Cannot inspect unavailable file when network access is disabled, package is locked or read-only.");
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"It is not possible to inspect a missing file when network access has been disabled in your preferences or the package is locked or not writable.\n"] attributes:[self signalAttributes]] autorelease]];
	    
	    [sheetDict setObject:self forKey:@"resulthandler"];
	    [sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
	    [sheetDict setObject:@"Alert" forKey:@"type"];
	    [sheetDict setObject:@"Warning" forKey:@"subtype"];
	    [sheetDict setObject:@"Disabled" forKey:@"title"];
	    [sheetDict setObject:@"It is not possible to inspect a missing file when network access has been disabled in your preferences or the package is locked or not writable." forKey:@"info"];
	    [sheetDict setObject:@"Accept" forKey:@"okbutton"];
	    [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
	    [__cachedWindowController autoButtonsEnabled];
	    [actionTask release];
	    return;
	}
	else {
	    II2TimestampComparisonResult compare;
	    [self readRemotePkgPlist:YES]; // Force read
	    compare = [self remoteStatus];
	    if (compare != NSOrderedSame) {
		[self compareSheet:compare forAction:@"inspect" withObject:fileName];
		[actionTask release];
		return;
	    }
	}
	if ([ii2Prefs autoFatten] || (modifiers&NSShiftKeyMask)) {
	    NSArray *tasksToLaunch = [self allocTasksForFileSetAvailable:useFileName forTask:actionTask filter:NO];
	    if (!tasksToLaunch) {
		NSLog( @"Could not get tasks for auto-downloading and checking file.");
		[actionTask release];
		[__cachedWindowController autoButtonsEnabled];
		return;
	    }
	    [actionTask launch];
	    [self markTaskArray:tasksToLaunch forAction:@"inspect"];
	    [self doActionTasks:tasksToLaunch];
	    [tasksToLaunch release];
	}
	else {
	    NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];

	    [sheetDict setObject:self forKey:@"resulthandler"];
	    [sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
	    [sheetDict setObject:@"Alert" forKey:@"type"];
	    [sheetDict setObject:@"Missing file" forKey:@"subtype"];
	    [sheetDict setObject:@"Missing file" forKey:@"title"];
	    [sheetDict setObject:[NSString stringWithFormat:@"%s (size: %@) is unavailable (try shift-click for automatic download of required files)", [useFileName fileSystemRepresentation], [[GWSizeFormatter sharedFormatter] stringForObjectValue:[NSNumber numberWithInt:[self sizeOfFile:useFileName]]]] forKey:@"info"];
	    [sheetDict setObject:@"Too bad" forKey:@"okbutton"];

	    [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
	}
	return;
    }
    else {
	NSMutableArray *tasksToLaunch = (NSMutableArray *)[self allocTasksForFileSetAvailable:useFileName forTask:actionTask filter:NO];
	if (!tasksToLaunch) {
	    NSLog( @"Some weird task allocation error happened when inspecting.");
	    return;
	}
	[tasksToLaunch addObject:actionTask];
	[self doActionTasks:tasksToLaunch];
	[tasksToLaunch release];
    }
}

- (void)showInfo
{
    NSString *infoFile = [__baseName stringByAppendingString:@".rtf"];

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: showInfo");
    if ([self fileIsPartOfPackage:infoFile]) {
	[self inspect:infoFile withModifiers:NSShiftKeyMask inView:[__cachedWindowController readmeView]];
    }
    else {
	[[__cachedWindowController readmeView] resetParagraphStyleAndClear];
	[[__cachedWindowController readmeView] appendOutput:@"This i-Package does not have a standard README RTF file."];
    }
}

- (BOOL)deleteFile:(NSString *)fileName
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: deleteFile %@", fileName);
    if ([self canBeFattened]) {
	BOOL isDir;
	NSString *fullPath = [self fullPathForFileName:fileName];
	NSString *inspectFileName = [self fileNameForInspect:fileName withModifiers:0];

	if (![fileName isEqualToString:inspectFileName]) {
	    if (![self deleteFile:inspectFileName]) {
		return NO;
	    }
	}

	if (![DEFFM fileExistsAtPath:fullPath isDirectory:&isDir]) {
	    return YES;
	}
	if (isDir) return NO;
	if ([ii2Prefs mainLogLevel] > 3) NSLog( @"Removing %s", [fullPath fileSystemRepresentation]);
	if (![DEFFM removeFileAtPath:fullPath handler:nil]) {
	    NSLog( @"Update: Unable to remove %s", [fullPath fileSystemRepresentation]);
	    return NO;
	}
	[__md5Checks removeObjectForKey:fileName];
	[__fileDataTypes removeObjectForKey:fileName];
	[__sigChecks removeObjectForKey:fileName];

	return YES;
    }
    else {
	NSLog( @"Cannot delete file when package is locked or not writable");
	return NO;
    }
}

- (void)gpgCheckToC
{
    NSArray *gpgTasks = [self gpgCheckTasks:[__baseName stringByAppendingString:@".plist"] forTask:nil];
    
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: gpgCheckToC");
    if ([gpgTasks count]) {
	[self doActionTasks:gpgTasks];
    }
    else {
	[__cachedWindowController autoButtonsEnabled];
    }
}

//<GLUPTaskOwnerProtocol> protocol and support:

- (GLUPTaskRunner *)md5CheckTask:(NSString *)fileName forTask:(id <II2TaskDependencyProtocol>)dependentTask
{
    NSArray *args = [NSArray arrayWithObjects: [II2TOOLFINDER pathForTool:@"/usr/bin/openssl"], @"md5", fileName, nil];
    GLUPTaskRunner *task;
    NSMutableData *outData = [[[NSMutableData alloc] init] autorelease];

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: md5CheckTask %@", fileName);
    task = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:args authenticator:nil];
    [task setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"md5check", @"type", outData, @"outdata", fileName, @"filename", nil]];
    [task setDirectoryPath:__dirName];
    [task setOutData:outData];
    if (dependentTask) {
	[task addDependentTask:dependentTask];
    }
    return task;
}

- (NSArray *)gpgCheckTasks:(NSString *)fileName forTask:(id <II2TaskDependencyProtocol>)dependentTask
{
    NSString *tmpDir = [NSString stringWithFormat:@"/tmp/.ii2gpg.%u.%u", getpid(), random()];
    NSString *testSigFile = [NSString stringWithFormat:@"%s/testsig", [tmpDir fileSystemRepresentation]];
    NSString *gpgHomeDir = [HOMEDIR stringByAppendingString:@"/Library/Caches/i-Installer/GPG"];
    BOOL isDir = NO;
    NSString *sig = nil;
    NSString *decompressor = @"/bin/cat";
    NSArray *decompressArgs, *gpgArgs;
    NSMutableData *errData; // GPG writes to stderr...
    GLUPTaskRunner *decompressTask, *gpgTask;
    BOOL checkingToC = [fileName isEqualToString:[__baseName stringByAppendingString:@".plist"]];

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: gpgCheckTasks %@", fileName);
    if (![DEFFM fileExistsAtPath:gpgHomeDir isDirectory:&isDir] || !isDir) {
	[DEFFM createDirectoryAtPath:[[gpgHomeDir stringByDeletingLastPathComponent] stringByDeletingLastPathComponent] attributes:nil];
	[DEFFM createDirectoryAtPath:[gpgHomeDir stringByDeletingLastPathComponent] attributes:nil];
	if (![DEFFM createDirectoryAtPath:gpgHomeDir attributes:[NSDictionary dictionaryWithObject:[NSNumber numberWithInt:0700] forKey:NSFilePosixPermissions]]) {
	    NSLog( @"Cannot create GPG Home Directory %@", gpgHomeDir);
	    return nil;
	}
    }
    if (![DEFFM createDirectoryAtPath:tmpDir attributes:nil]) {
	NSLog( @"Cannot create GPG Task temporary directory %@", tmpDir);
	return nil;
    }
    if (checkingToC) {
	NSString *plistSigFile = [self fullPathForFileName:[__baseName stringByAppendingString:@".plist.sig"]];
	NSString *sigContents = [NSString stringWithContentsOfFile:plistSigFile];

	if (![sigContents length]) {
	    [DEFFM removeFileAtPath:tmpDir handler:nil];
	    NSLog( @"Serious problem: could not remove tmeporary directory for GPG.");
	    return nil;
	}
	[sigContents writeToFile:testSigFile atomically:NO];
    }
    else {
	NSDictionary *fileDict = [__pkgPlist objectForKey:@"files"];
	NSDictionary *file = [fileDict objectForKey:fileName];
	sig = [file objectForKey:@"sig"];
	if (![sig length]) {
	    NSLog( @"Serious problem: remote signature missing in local plist.");
	    [DEFFM removeFileAtPath:tmpDir handler:nil];
	    return nil;
	}
	[sig writeToFile:testSigFile atomically:NO];
    }
    errData = [[[NSMutableData alloc] init] autorelease];
    decompressArgs = [NSArray arrayWithObjects: decompressor, fileName, nil];
    decompressTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:decompressArgs authenticator:nil];
    gpgArgs = [NSArray arrayWithObjects: [[NSBundle mainBundle] pathForResource:@"gpg" ofType:nil], ([ii2Prefs mainLogLevel] > 4 ? @"--verbose" : @"--quiet"), @"--keyserver", [ii2Prefs gpgKeyServer], @"--homedir", gpgHomeDir, @"--no-options", @"--keyserver-options", @"auto-key-retrieve", @"--verify", testSigFile, @"-", nil];
    gpgTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:gpgArgs authenticator:nil];
    [decompressTask setDirectoryPath:__dirName];
    [decompressTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"ignore", @"type", nil]];
    [gpgTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: (checkingToC ? @"tocsigcheck" : @"sigcheck"), @"type", fileName, @"filename", errData, @"outdata", nil]];

    [gpgTask useTmpDir:tmpDir cleanOnFinish:([ii2Prefs mainLogLevel] < 4)];
    [gpgTask setErrData:errData]; // GPG writes on std err!
    [decompressTask connectOutputTo:gpgTask];
    if (dependentTask) {
	[gpgTask addDependentTask:dependentTask];
    }
    return [[[NSArray alloc] initWithObjects: decompressTask, gpgTask, nil] autorelease];
}

- (NSArray *)dataTypeCheckTasks:(NSString *)fileName forTask:(id <II2TaskDependencyProtocol>)dependentTask
{
    NSString *decompressor = @"/bin/cat";
    NSArray *decompressArgs, *fileArgs;
    GLUPTaskRunner *decompressTask, *fileTask;
    NSMutableData *outData = [[[NSMutableData alloc] init] autorelease];

    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: dataTypeCheckTasks %@", fileName);
    if ([fileName hasSuffix:@".gz"] || [fileName hasSuffix:@".bz2"]) {
	decompressor = [fileName hasSuffix:@".gz"] ? [II2TOOLFINDER pathForTool:@"/usr/bin/gzcat"] : [II2TOOLFINDER pathForTool:@"/usr/bin/bzcat"];
    }
    decompressArgs = [NSArray arrayWithObjects: decompressor, fileName, nil];
    decompressTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:decompressArgs authenticator:nil];
    fileArgs = [NSArray arrayWithObjects: [[NSBundle mainBundle] pathForResource:@"file" ofType:nil], @"-m", @"magic", @"-", nil];
    fileTask = [[GLUPTaskRunner alloc] initWithOwner:self reportsPID:GLUPTaskRunnerFifoPid arguments:fileArgs authenticator:nil];
    [fileTask setDirectoryPath:[[NSBundle mainBundle] resourcePath]]; // to be able to find magic

    [decompressTask setDirectoryPath:__dirName];
    [decompressTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"ignore", @"type", nil]];
    [fileTask setProperties:[NSDictionary dictionaryWithObjectsAndKeys: @"filecheck", @"type", outData, @"outdata", fileName, @"filename", nil]];

    [fileTask setOutData:outData];
    [decompressTask connectOutputTo:fileTask];
    if (dependentTask) {
	[fileTask addDependentTask:dependentTask];
    }
    return [[[NSArray alloc] initWithObjects: decompressTask, fileTask, nil] autorelease];
}

- (void)unarchiveSentinel
{
    NSArray *selectedSets = [self selectedKnownSets];
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: unarchiveSentinel");
    if (selectedSets) {
	// Clear configured state and set installed state for sets we have just unarchived
	int i;
	
	for (i=0; i<[selectedSets count]; i++) {
	    NSString *setId = [[selectedSets objectAtIndex:i] objectForKey:@"id"];
	    [self setSelectorSet:setId configured:NO];	    
	    [self setSelectorSet:setId installed:YES];	    
	}
    }
    [self writeReceipt];
    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Unarchive stage completed\n"] attributes:[self infoAttributes]] autorelease]];
    if (![self fileIsPartOfPackage:[__baseName stringByAppendingString:@".configure"]]) {
        NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];
        
        // Unarchive not followed by configure: we are done
        [sheetDict setObject:self forKey:@"resulthandler"];
        [sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
        [sheetDict setObject:@"Alert" forKey:@"type"];
        [sheetDict setObject:@"Task finished" forKey:@"subtype"];
        [sheetDict setObject:@"Task finished" forKey:@"title"];
        [sheetDict setObject:@"Installation completed succesfully." forKey:@"info"];
        [sheetDict setObject:@"OK" forKey:@"okbutton"];
        
        [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
    }
}

- (void)glupTaskDidFinish: (GLUPTaskRunner *)task withSuccess:(BOOL)success
{
    NSString *taskType;
    BOOL taskAborted;
    NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];

    if ([ii2Prefs mainLogLevel] > 1) NSLog( @"GLUP Task (%@) did finish with success %u, retainCount = %u", [task description], success, [task retainCount]);
    // Handle results if necessary

    // Basic setttings for sheetDict
    [sheetDict setObject:self forKey:@"resulthandler"];
    [sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
    [sheetDict setObject:@"Alert" forKey:@"type"];
    [sheetDict setObject:@"Task failed" forKey:@"subtype"];
    [sheetDict setObject:@"Task failed" forKey:@"title"];

    // [__ii2RunningTaskSet ii2TaskDidEnd:task withSuccess:([task result] == 0)];
    taskAborted = [[task propertyForKey:@"aborted"] isEqualToString:@"YES"];
    taskType = [[task properties] objectForKey:@"type"];
    if ([taskType isEqualToString:@"md5check"]) {
	NSString *fileName = [[task properties] objectForKey:@"filename"];
	if (success) {
	    // Get result from outData
	    NSDictionary *fileDict = [__pkgPlist objectForKey:@"files"];
	    NSDictionary *file = [fileDict objectForKey:fileName];
	    NSString *md5 = [file objectForKey:@"md5"];
	    NSData *resultData = [[task properties] objectForKey:@"outdata"];
	    NSString *resultString = [[[NSString alloc] initWithData:resultData encoding:NSASCIIStringEncoding] autorelease];
	    if (md5StringsAreEqual( md5, [resultString substringToIndex:[resultString length]-1])) {
		[__md5Checks setObject: @"YES" forKey:fileName];
		if ([ii2Prefs mainLogLevel] > 5) NSLog( @"Valid MD5 hash for file: %s", [fileName fileSystemRepresentation]);
	    }
	    else {
		NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];

		NSLog( @"Invalid MD5 hash for file: %s", [fileName fileSystemRepresentation]);
		[[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"MD5 Integrity check for file: %s failed (badly behaving cache/proxy?)\n", [fileName fileSystemRepresentation]] attributes:[self signalAttributes]] autorelease]];
		
		[__md5Checks setObject: @"NO" forKey:fileName];

		[sheetDict setObject:self forKey:@"resulthandler"];
		[sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
		[sheetDict setObject:@"Alert" forKey:@"type"];
		[sheetDict setObject:@"Warning" forKey:@"subtype"];
		[sheetDict setObject:@"MD5 intergrity check failed!" forKey:@"title"];
		[sheetDict setObject:@"i-Installer detected an file that does not match the MD5 integrity checksum in your package's Table of Contents. Normally this implies a bad behaving proxy/cache between you and the internet." forKey:@"info"];
		[sheetDict setObject:@"Accept" forKey:@"okbutton"];
		[[__cachedWindowController interactionManager] startInternalPanel:sheetDict];

		if (![ii2Prefs keepInvalidMd5]) {
		    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"%s will be deleted.\n", [fileName fileSystemRepresentation]] attributes:[self signalAttributes]] autorelease]];
		    [self deleteFile:fileName];
		}
	    }
	    [__cachedWindowController refreshFileTable]; // noop if file not in that particular table
	}
	else {
	    NSLog( @"MD5 check for file: %s failed", [fileName fileSystemRepresentation]);
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"MD5 Integrity check for file: %s failed (task aborted)\n", [fileName fileSystemRepresentation]] attributes:[self signalAttributes]] autorelease]];
	}
    }
    else if ([taskType isEqualToString:@"sigcheck"] || [taskType isEqualToString:@"tocsigcheck"]) {
	NSString *fileName = [[task properties] objectForKey:@"filename"];
	NSData *resultData = [[task properties] objectForKey:@"outdata"];
	NSString *resultString = [[[NSString alloc] initWithData:resultData encoding:NSASCIIStringEncoding] autorelease];
	if (success) {
	    if ([taskType isEqualToString:@"sigcheck"]) {
		[__sigChecks setObject: @"YES" forKey:fileName];
	    }
	    if ([ii2Prefs mainLogLevel] > 5) NSLog( @"GPG Sig check for file: %s succeeded", [fileName fileSystemRepresentation]);
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Succesful GPG Signature Check for file %s\n", [fileName fileSystemRepresentation]] attributes:[self infoAttributes]] autorelease]];
	}
	else {
	    if ([taskType isEqualToString:@"sigcheck"]) {
		[__sigChecks setObject: @"NO" forKey:fileName];
	    }
	    else {
		// tocsigcheck unsuccesful. Danger!
		NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];
		
		NSLog( @"i-Installer detected an invalid GPG signature on your package's Table of Contents. This is a sign of tampering, but it also could reflect problems with GPG. If you trust the package, you can go ahead, but this is not recommended.");
		[[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"GPG Signature check for table of contents failed (badly behaving cache/proxy?)\n", [fileName fileSystemRepresentation]] attributes:[self signalAttributes]] autorelease]];
		
		[sheetDict setObject:self forKey:@"resulthandler"];
		[sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
		[sheetDict setObject:@"Alert" forKey:@"type"];
		[sheetDict setObject:@"Warning" forKey:@"subtype"];
		[sheetDict setObject:@"Danger: Invalid GPG signature!" forKey:@"title"];
		[sheetDict setObject:@"i-Installer detected an invalid GPG signature on your package's Table of Contents. This is a sign of tampering, but it also could reflect problems with GPG. If you trust the package, you can go ahead, but this is not recommended." forKey:@"info"];
		[sheetDict setObject:@"I Understand" forKey:@"okbutton"];
		[[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
	    }
	    if ([ii2Prefs mainLogLevel] > 0) NSLog( @"GPG Sig check for file: %s failed", [fileName fileSystemRepresentation]);
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"GPG Signature check for file: %s failed\n", [fileName fileSystemRepresentation]] attributes:[self signalAttributes]] autorelease]];
	}
	[[__cachedWindowController signatureView] appendAttributedOutput: [[[NSAttributedString alloc] initWithString:resultString attributes:[[__cachedWindowController outView] fixedNormalAttributes]] autorelease]];
	[__cachedWindowController refreshFileTable]; // noop if file not in that particular table
    }
    else if ([taskType isEqualToString:@"rtfinspect"]) {
	if (success) {
	    NSData *resultData = [[task properties] objectForKey:@"outdata"];
	    [[[task properties] objectForKey:@"view"] selectAll:nil];
	    [[[task properties] objectForKey:@"view"] replaceCharactersInRange:[[[task properties] objectForKey:@"view"] selectedRange] withRTF:resultData];
	    if ([[task properties] objectForKey:@"view"] == [__cachedWindowController readmeView]) {
		[__cachedWindowController showTab:ii2ReadmeTabTag];
	    }
	    else if ([[task properties] objectForKey:@"view"] == [__cachedWindowController inspectView]) {
		[__cachedWindowController showTab:ii2ContentsTabTag];
	    }
	}
    }
    else if ([taskType isEqualToString:@"filecheck"]) {
	if (success) {
	    NSString *fileName = [[task properties] objectForKey:@"filename"];
	    NSData *resultData = [[task properties] objectForKey:@"outdata"];
	    NSString *resultString = [[[NSString alloc] initWithData:resultData encoding:NSASCIIStringEncoding] autorelease];
	    NSString *chompedString = [resultString substringWithRange:NSMakeRange(29,[resultString length]-30)];
	    [__fileDataTypes setObject:chompedString forKey:fileName];
	    [__cachedWindowController refreshFileTable]; // noop if file not in that particular table
	}
    }
    else if ([taskType isEqualToString:@"installstamp"] || [taskType isEqualToString:@"removestamp"]) {
	if (success) {
	    [__cachedWindowController updatePropertiesDisplay];
	}
    }
    else if ([taskType isEqualToString:@"prepare"]) {
	if (!success) {
	    if (taskAborted) {
		[sheetDict setObject:@"Preparation for install was aborted." forKey:@"info"];
		[sheetDict setObject:@"OK" forKey:@"okbutton"];
	    }
	    else {
		[sheetDict setObject:@"The prepare stage of the install ended with a failure. See output view for possible clues." forKey:@"info"];
		[sheetDict setObject:@"Accept" forKey:@"okbutton"];
	    }

	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Prepare stage failed\n"] attributes:[self signalAttributes]] autorelease]];
	    [__cachedWindowController stopDOServer];
	    [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
	}
	else {
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Prepare stage completed\n"] attributes:[self infoAttributes]] autorelease]];
	}
    }
    else if ([taskType isEqualToString:@"unarchivepart"]) {
	if (!success) {
	    [self removeReceipt];
	    if (taskAborted) {
		[sheetDict setObject:@"Unarchiving was aborted." forKey:@"info"];
		[sheetDict setObject:@"OK" forKey:@"okbutton"];
	    }
	    else {
		[sheetDict setObject:@"Unarchiving ended with a failure. See output view for possible clues." forKey:@"info"];
		[sheetDict setObject:@"Accept" forKey:@"okbutton"];
	    }

	    [__cachedWindowController stopDOServer];
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Unarchive stage failed\n"] attributes:[self signalAttributes]] autorelease]];
	    [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
	}
	else {
	    if ([taskType isEqualToString:@"unarchivepart"]) {
                [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Unarchive stage archive fragment %@ %@ in directory %@\n", [[task properties] objectForKey:@"archivepartnumber"], ([task skip] ? @"skipped" : @"completed"), [[task properties] objectForKey:@"partinstalllocation"]] attributes:[self infoAttributes]] autorelease]];
	    }
        }
    }
    else if ([taskType isEqualToString:@"configure"]) {
	NSString *userAction = [[task properties] objectForKey:@"useraction"];
	[__cachedWindowController stopDOServer];
	if (!success) {
	    if (taskAborted) {
		[sheetDict setObject:@"The configure script was aborted." forKey:@"info"];
		[sheetDict setObject:@"OK" forKey:@"okbutton"];
	    }
	    else {
		[sheetDict setObject:@"Configuration ended with a failure. See output view for possible clues." forKey:@"info"];
		[sheetDict setObject:@"Accept" forKey:@"okbutton"];
	    }
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Configure stage failed\n"] attributes:[self signalAttributes]] autorelease]];
	    [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
	}
	else {
	    [self writeReceipt]; // Save the changes we made during configuration
            [__cachedWindowController restoreTabViewSettings];
	    [sheetDict setObject:@"Task finished" forKey:@"subtype"];
	    [sheetDict setObject:@"Task finished" forKey:@"title"];
	    if (userAction && [userAction isEqualToString:@"trueinstall"]) {
		[sheetDict setObject:@"Installation and running the configuration script finished without critical errors." forKey:@"info"];
	    }
	    else {
		[sheetDict setObject:@"Running the configuration script finished without critical errors." forKey:@"info"];
	    }
	    [sheetDict setObject:@"OK" forKey:@"okbutton"];

	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Configure stage completed\n"] attributes:[self infoAttributes]] autorelease]];
	    [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
	}
    }
    else if ([taskType isEqualToString:@"remove"]) {
	[__cachedWindowController stopDOServer];
	if (!success) {
	    if (taskAborted) {
		[sheetDict setObject:@"The remove script was aborted." forKey:@"info"];
		[sheetDict setObject:@"OK" forKey:@"okbutton"];
	    }
	    else {
		[sheetDict setObject:@"The remove script ended with a failure. See output view for possible clues." forKey:@"info"];
		[sheetDict setObject:@"Accept" forKey:@"okbutton"];
	    }
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Remove stage failed\n"] attributes:[self infoAttributes]] autorelease]];
	    [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
	}
	else {
	    [self removeReceipt];
            [__cachedWindowController restoreTabViewSettings];
	    [sheetDict setObject:@"Task finished" forKey:@"subtype"];
	    [sheetDict setObject:@"Task finished" forKey:@"title"];
	    [sheetDict setObject:@"The remove script finished without critical errors." forKey:@"info"];
	    [sheetDict setObject:@"OK" forKey:@"okbutton"];

	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Remove stage completed\n"] attributes:[self infoAttributes]] autorelease]];
	    [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
	}
    }
    else if ([taskType isEqualToString:@"selector"]) {
	[__cachedWindowController stopDOServer];
	if (!success) {
	    if (taskAborted) {
		[sheetDict setObject:@"The selector script was aborted." forKey:@"info"];
		[sheetDict setObject:@"OK" forKey:@"okbutton"];
	    }
	    else {
		[sheetDict setObject:@"The selector script ended with a failure. See output view for possible clues." forKey:@"info"];
		[sheetDict setObject:@"Accept" forKey:@"okbutton"];
	    }
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Selector stage failed\n"] attributes:[self signalAttributes]] autorelease]];
	    [[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
	}
	else {
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Selector stage completed\n"] attributes:[self infoAttributes]] autorelease]];
	    [self action:@"dependencycheckandinstall"];
	}
    }
    else {
	if ([ii2Prefs mainLogLevel] > 4) NSLog( @"Task %@ for action %@ signalled completion (unhandled)", [task description], taskType);
    }
}

- (void)glupTaskWillStart: (GLUPTaskRunner *)task
{
    NSString *taskType = [[task properties] objectForKey:@"type"];
    NSString *stdoutMessage = [[task properties] objectForKey:@"stdoutmessage"];

    if ([ii2Prefs mainLogLevel] >1) NSLog( @"GLUP Task (%@) will start, retainCount = %u", [task description], [task retainCount]);
    if ([ii2Prefs dynamicDisplayOfTabs] && ([taskType isEqualToString:@"selector"] || [taskType isEqualToString:@"prepare"] || [taskType isEqualToString:@"unarchivepart"] || [taskType isEqualToString:@"configure"])) {
        [__cachedWindowController showTab:ii2OutputTabTag];
    }
    if (stdoutMessage) {
	[[__cachedWindowController inspectView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:stdoutMessage attributes:[self infoAttributes]] autorelease]];
    }
}

- (void)downloadTaskDidFinish: (II2DownloadTask *)task withSuccess:(BOOL)success
{
    NSString *fileName = [[task properties] objectForKey:@"filename"];

    if ([ii2Prefs mainLogLevel] >1) NSLog( @"Download Task (%@) did finish, retainCount = %u", [task description], [task retainCount]);
    if (success) {
	NSDictionary *fileDict = [__pkgPlist objectForKey:@"files"];
	NSDictionary *file = [fileDict objectForKey:fileName];
	NSString *executable = [file objectForKey:@"executable"];
	if (executable && [executable isEqualToString:@"1"]) {
	    NSDictionary *permDict = [NSDictionary dictionaryWithObject:[NSNumber numberWithInt:0755] forKey:NSFilePosixPermissions];
	    if (![DEFFM changeFileAttributes:permDict atPath:[task fullPath]]) {
		NSLog( @"Failed to set executable file attributes for downloaded file %s", [[task fullPath] fileSystemRepresentation]);
	    }
	}
    }
    [self clearChecksForFile:fileName];
    [__ii2RunningTaskSet refreshDisplay:task];
}

- (void)downloadTaskWillStart: (II2DownloadTask *)task
{
    if ([ii2Prefs mainLogLevel] >1) NSLog( @"Download Task (%@) will start, retainCount = %u", [task description], [task retainCount]);
    if ([ii2Prefs dynamicDisplayOfTabs]) {
        [__cachedWindowController showTab:ii2ActivityTabTag];
    }
}

- (void)actionTaskDidFinish: (II2ActionTask *)task withSuccess:(BOOL)success
{
    if ([ii2Prefs mainLogLevel] >1) NSLog( @"Action Task (%@) did finish, retainCount = %u", [task description], [task retainCount]);
    [__ii2RunningTaskSet refreshDisplay:task];
    //if ([[task description] isEqualToString:@"__inspect:inView:"]) {
	[__cachedWindowController autoButtonsEnabled];
    //}
}

- (void)actionTaskWillStart: (II2ActionTask *)task
{
    if ([ii2Prefs mainLogLevel] >1) NSLog( @"Action Task (%@) will start, retainCount = %u", [task description], [task retainCount]);
    if ([ii2Prefs dynamicDisplayOfTabs]) {
        [__cachedWindowController showTab:ii2ActivityTabTag];
    }
}

- (BOOL)ii2TaskObject: (id)object1 isEqualTo: (id)object2
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: ii2TaskObject");
    return [(NSString *)object1 isEqualToString:(NSString *)object2];
}

- (void)abortTaskAtIndex:(int)index
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: abortTaskAtIndex");
    if ((index >= 0) && (index < [__ii2RunningTaskSet count])) {
	id task = [__ii2RunningTaskSet taskAtIndex:index];

        if ([ii2Prefs mainLogLevel] >1) NSLog( @"Task (%@) will abort, retainCount = %u", [task description], [task retainCount]);
	[task abort];
    }
}

- (void)fatten
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: fatten");
    if (![self canBeFattened]) {
	NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];
	
	NSLog( @"It is not possible to download missing parts of packages when network access has been disabled in your preferences, or the package is locked or not writable. Check your preferences or lock status or copy the package to a writable location first.");
	
	[sheetDict setObject:self forKey:@"resulthandler"];
	[sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
	[sheetDict setObject:@"Alert" forKey:@"type"];
	[sheetDict setObject:@"Warning" forKey:@"subtype"];
	[sheetDict setObject:@"Cannot Fatten" forKey:@"title"];
	[sheetDict setObject:@"It is not possible to download missing parts of packages when network access has been disabled in your preferences, or the package is locked or not writable. Check your preferences or lock status or copy the package to a writable location first." forKey:@"info"];
	[sheetDict setObject:@"Accept" forKey:@"okbutton"];
	[[__cachedWindowController interactionManager] startInternalPanel:sheetDict];
    }
    else {
	NSArray *fattenTasks = [self allocTasksForAction:@"fatten"];
	if ([self checkForDownloadsInTaskArray:fattenTasks]) {
	    II2TimestampComparisonResult compare;
	    [self readRemotePkgPlist:YES]; // Force read
	    compare = [self remoteStatus];
	    if (compare != NSOrderedSame) {
		NSMutableDictionary *sheetDict = [[[NSMutableDictionary alloc] init] autorelease];
		
		NSLog( @"Fattening aborted. remote package is not the same version.");
		
		[sheetDict setObject:self forKey:@"resulthandler"];
		[sheetDict setObject:NSStringFromSelector( @selector(endSheetWithDictionary:)) forKey:@"resulthandlerselector"];
		[sheetDict setObject:@"Alert" forKey:@"type"];
		[sheetDict setObject:@"Warning" forKey:@"subtype"];
		[sheetDict setObject:@"Abort Fatten" forKey:@"title"];
		[sheetDict setObject:@"Fattening was aborted because the remote package version is not the same as this package version and downloading would thus automatically fail for changed items." forKey:@"info"];
		[sheetDict setObject:@"Accept" forKey:@"okbutton"];
		[[__cachedWindowController interactionManager] startInternalPanel:sheetDict];

		[[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Fattening aborted. remote package is not the same version.\n"] attributes:[self signalAttributes]] autorelease]];
		[self destroyTaskArray:fattenTasks];
		return;
	    }
	}
	else {
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"No downloads were necessary, package is already fat. Checking intergrity of files only...\n"] attributes:[self infoAttributes]] autorelease]];
	}
	[self doActionTasks:fattenTasks];
	[fattenTasks release];
    }
}

- (void)thin
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: thin");
    if (![self isLocked] && [self isWritable]) {
	NSMutableArray *existingFileArray = [[[NSMutableArray alloc] init] autorelease];
	NSEnumerator *enumerator = [__completeFileArray objectEnumerator];
	NSDictionary *fileDict;
	unsigned int totalSize = 0;
	unsigned int thinSize = [ii2Prefs thinSize];

	NSString *plistName = [NSString stringWithFormat:@"%@%@.plist", __dirNameWithSeparator, __baseName];

	while ((fileDict = [enumerator nextObject]) != nil) {
	    NSString *fileName = [fileDict objectForKey:@"file"];
	    if (![plistName isEqualToString:fileName] && [self fileIsAvailable:fileName]) {
		[existingFileArray addObject:fileDict];
		totalSize += [self sizeOfFile:fileName];
	    }
	}
	[existingFileArray sortUsingFunction:compareFileDictOnSize context:nil];

	enumerator = [existingFileArray objectEnumerator];
	while ((totalSize > thinSize) && ((fileDict = [enumerator nextObject]) != nil)) {
	    NSString *fileName = [fileDict objectForKey:@"file"];
	    if (![self deleteFile:fileName]) {
		return;
	    }
	    totalSize -= [self sizeOfFile:fileName];
	    [[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"File %@ removed from package (size of file is %@).\n", fileName, [[GWSizeFormatter sharedFormatter] stringForObjectValue:[NSNumber numberWithInt:[self sizeOfFile:fileName]]]] attributes:[self infoAttributes]] autorelease]];
	}
	[[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Package has been thinned. The package now uses %@ on disk.\n", [[GWSizeFormatter sharedFormatter] stringForObjectValue:[NSNumber numberWithInt:totalSize]]] attributes:[self infoAttributes]] autorelease]];
    }
    else {
	[[__cachedWindowController messageView] appendAttributedOutput:[[[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"Package cannot be thinned because it is locked, or on not writable.\n"] attributes:[self signalAttributes]] autorelease]];
    }
}

- (BOOL)active
{
    if ([ii2Prefs mainLogLevel] > 9) NSLog( @"II2P: active");
    return [__ii2RunningTaskSet count] != 0;
}

- (void) reportRCSIDActions
{
    NSLog( @"%s\n", rcsidActions);
}

@end

/*
 * $Log: II2PackageActions.m,v $
 * Revision 1.11  2004/10/11 20:16:24  gctwnl
 * Version 2.66
 *
 * Revision 1.10  2004/05/01 12:31:34  gctwnl
 * 2.63.0
 *
 * Revision 1.9  2004/04/30 12:18:06  gctwnl
 * 2.62.0
 *
 * Revision 1.8  2004/04/25 13:13:21  gctwnl
 * Forced commit of all .h and .m files of version 2.61.0
 *
 * Revision 1.7  2004/03/21 16:52:21  gctwnl
 * 2.58.0
 *
 * Revision 1.6  2004/03/20 22:14:52  gctwnl
 * 2.57.0
 *
 * Revision 1.5  2004/03/20 14:04:48  gctwnl
 * 2.56.0
 *
 * Revision 1.4  2004/03/19 23:35:11  gctwnl
 * Version 2.55.0. Has a renewed interface to support simple operations.
 * Supports progress bar control from the i-Package.
 *
 * Revision 1.3  2004/03/12 09:50:29  gctwnl
 * Version 2.53.0
 *
 * Revision 1.2  2004/02/27 23:24:27  gctwnl
 * 2.52.0 entry after a long period of CVS problems
 *
 * Revision 1.1.1.1  2004/02/01 17:18:25  gctwnl
 * Initial import.
 *
 * Revision 3.27  2004/01/31 16:13:43  gerben
 * 2.49.0. Bug fix dependency checking
 *
 * Revision 3.26  2004/01/27 22:49:32  gerben
 * 2.45.0
 *
 * Revision 3.25  2004/01/27 12:57:39  gerben
 * 2.44.0
 *
 * Revision 3.24  2004/01/23 00:31:16  gerben
 * 2.42.0 release
 *
 * Revision 3.23  2004/01/15 01:00:06  gerben
 * 2.40.0
 *
 * Revision 3.22  2004/01/14 12:56:46  gerben
 * Version 2.39.0
 *
 * Revision 3.21  2004/01/05 23:47:15  gerben
 * *** empty log message ***
 *
 * Revision 3.20  2004/01/05 15:31:34  gerben
 * *** empty log message ***
 *
 * Revision 3.19  2003/12/05 12:40:52  gerben
 * Fixed .stamp and Custom.plist handling. Changed GPG key server default.
 *
 * Revision 3.18  2003/11/09 13:02:07  gerben
 * 2.31.0
 *
 * Revision 3.17  2003/11/08 23:38:01  gerben
 * 2.30.0 release candidate
 *
 * Revision 3.16  2003/09/22 15:57:38  gerben
 * 2.27. Fixed bug wrt to spaces in directories when unarchiving through sh
 *
 * Revision 3.15  2003/09/18 22:40:54  gerben
 * 2.26.0 Fragmented archives
 *
 * Revision 3.14  2003/08/03 18:52:35  gerben
 * Added ZIP handling. Version 2.25.0.
 *
 * Revision 3.13  2003/05/01 13:45:28  gerben
 * 2.20.1. Added messages to thinning process.
 *
 * Revision 3.12  2003/04/28 08:05:26  gerben
 * 2.20.0. iibless improvements by Tom Hageman, a way for remove scripts to communicate non-removal back to i-Installer.
 *
 * Revision 3.11  2003/03/16 22:30:44  gerben
 * Added environment variables for PKG time stamp to subprocess environment. These show up in a report.
 *
 * Revision 3.10  2003/03/14 10:10:19  gerben
 * 2.13.0. Update resets Inspect view
 *
 * Revision 3.9  2003/03/13 16:15:55  gerben
 * 2.12.0, go back on the permissions problem. The permissions need to be in the archive anyway, because root extraction overrides everything.
 *
 * Revision 3.8  2003/03/13 14:42:04  gerben
 * 2.11.0. Really fix the permissions problem.
 *
 * Revision 3.7  2003/03/11 22:23:33  gerben
 * Added warning when package cannot be created. Removed p flag from gnutar, unarchived files will be owned by root or user depending on authentication
 *
 * Revision 3.6  2003/02/23 23:29:11  gerben
 * Package Locking
 *
 * Revision 3.5  2003/02/10 16:03:35  gerben
 * 2.6.0 Fixed horrible crasher bug introduced when fattening.
 *
 * Revision 3.4  2003/02/10 14:07:46  gerben
 * Fixed a bug where a task array was not destroyed with destroyTaskArray (which creates a file descriptor leak)
 *
 * Revision 3.3  2002/12/16 13:13:38  gerben
 * Bug fix: The install directory was not created when one chose to install while ignoring the remote availability of another version.
 *
 * Revision 3.2  2002/12/10 14:24:52  gerben
 * Set sticky tag 3.2
 *
 * Revision 3.0  2002/12/10 14:04:23  gerben
 * Upping all files with a sticky tag of 1.99
 *
 * Revision 1.99  2002/12/10 13:52:52  gerben
 * Set all revision numbers to 1.99
 *
 * Revision 1.80  2002/12/10 13:50:17  gerben
 * A bit of cleanup. One bugfix (wrong method)
 *
 * Revision 1.79  2002/12/10 08:55:25  gerben
 * II2Package code cleanup, remove of cached view pointers. Added relocatable handling in package dependencies.
 *
 * Revision 1.78  2002/12/09 15:08:59  gerben
 * Report printing
 *
 * Revision 1.77  2002/12/03 08:16:57  gerben
 * New style install stamps .foo.ii2
 *
 * Revision 1.76  2002/12/02 23:24:41  gerben
 * Basic dependency handling
 *
 * Revision 1.75  2002/11/30 10:05:55  gerben
 * Rearranged code for II2Package to use catagories instead of code inclusion.
 *
 * Revision 1.74  2002/11/27 23:43:31  gerben
 * Fixed bug where tasks not executed were never deallocated. This led to a fd leak....
 *
 * Revision 1.73  2002/11/24 23:09:27  gerben
 * Tom's changes and remarks mainly
 *
 * Revision 1.72  2002/11/22 23:54:47  gerben
 * RC4
 *
 * Revision 1.71  2002/10/29 09:36:22  gerben
 * RC2
 *
 * Revision 1.70  2002/10/27 15:39:43  gerben
 * Tried to remove BSD-package dependency
 *
 * Revision 1.69  2002/10/27 12:38:34  gerben
 * Lazy update checking. Update followed by action for inspect and fatten.
 *
 * Revision 1.68  2002/10/27 11:19:50  gerben
 * Sig checking changed from remote to local for TOC. Update does not update when timestamps inside TOCs are equal.
 *
 * Revision 1.67  2002/10/25 20:36:11  gerben
 * Layout changes.
 *
 * Revision 1.66  2002/10/25 19:52:35  gerben
 * Changes to update reporting.
 *
 * Revision 1.65  2002/10/25 19:27:22  gerben
 * Several small changes re known packages window. Implemented developer build stamp in main bundle.
 *
 * Revision 1.64  2002/10/24 20:12:32  gerben
 * Better handling of package name clashes and changed properties
 *
 * Revision 1.63  2002/10/24 13:03:31  gerben
 * Disable menu items that can affect downloads and availability of files when package is active.
 *
 * Revision 1.62  2002/10/24 12:49:05  gerben
 * Update refreshes file table. Download reports when remainder of (not total) download time changes significantly.
 *
 * Revision 1.61  2002/10/22 22:20:58  gerben
 * Small changes
 *
 * Revision 1.60  2002/10/22 14:14:19  gerben
 * Differ between different user actions.
 *
 * Revision 1.59  2002/10/22 12:52:20  gerben
 * Added stderr handling to authenticated tasks.
 *
 * Revision 1.58  2002/10/21 20:16:53  gerben
 * Closing in on RC1 target
 *
 * Revision 1.57  2002/10/17 18:48:59  gerben
 * Cache/NoCache handling on downloads. Some other changes.
 *
 * Revision 1.56  2002/10/16 22:37:07  gerben
 * A few small changes.
 *
 * Revision 1.55  2002/10/16 21:28:21  gerben
 * A complete new rewrite of taskss, processes and stuff.
 *
 * Revision 1.54  2002/10/16 08:37:18  gerben
 * Changed log levels. Small fixes.
 *
 * Revision 1.53  2002/10/15 14:31:44  gerben
 * Last fixes. No-network access. Later creation of downloaded file.
 *
 * Revision 1.52  2002/10/15 10:31:18  gerben
 * A task array was released while actions were already running. Reason: we started the tasks without authenticating first, so authentication came when we were already started. TODO: this can happen too when authorizations expire!
 *
 * Revision 1.51  2002/10/14 12:42:56  gerben
 * Some name changes
 *
 * Revision 1.50  2002/10/14 12:28:20  gerben
 * Some extra progress messages
 *
 * Revision 1.49  2002/10/14 12:15:37  gerben
 * Install stamp file after unarchive.
 *
 * Revision 1.48  2002/10/13 22:22:27  gerben
 * Re-created file menu so validating menu items work. Changed code for inspect and fatten to use generic task array starter.
 *
 * Revision 1.47  2002/10/13 21:20:14  gerben
 * Fixed a serious bug where the task rewrite was a bit release-happy
 *
 * Revision 1.46  2002/10/12 10:54:38  gerben
 * A lot of changes, mostly when producing output/feedback.
 *
 * Revision 1.45  2002/10/10 00:51:24  gerben
 * Many changes. Committing a stable state.
 *
 * Revision 1.44  2002/10/09 14:22:23  gerben
 * Removed outputHandler from GLUPTaskRunner
 *
 * Revision 1.43  2002/10/09 14:16:11  gerben
 * Changed as many %s/lossyCString uses to %@/fileSystemRepresentation/UTF8String
 *
 * Revision 1.42  2002/10/09 13:33:56  gerben
 * Success message on completion of install without configure.
 *
 * Revision 1.41  2002/10/09 10:53:40  gerben
 * Moved textview output to NSTexTView category.
 *
 * Revision 1.40  2002/10/09 08:38:42  gerben
 * Fixed two 10.1.5 incompatibilities (untested)
 *
 * Revision 1.39  2002/10/08 22:25:02  gerben
 * GPG checking complete. Output of parallel tasks will not mingle.
 *
 * Revision 1.38  2002/10/08 21:49:48  gerben
 * GPG sigs checked when preferences change.
 *
 * Revision 1.37  2002/10/08 21:34:28  gerben
 * Working GPG signature checks, shared GPG cache for II2.
 *
 * Revision 1.36  2002/10/08 13:20:29  gerben
 * MD5 check on update. Small changes.
 *
 * Revision 1.35  2002/10/08 10:03:42  gerben
 * Several small changes, amongst others first responder behaviour.
 *
 * Revision 1.34  2002/10/07 20:28:05  gerben
 * Removed circular CVS Log entry in C Comments at the end of files
 *
 * Revision 1.33  2002/10/07 20:20:03  gerben
 * More md5 stuff and testing of spinning disk progressindicator stuff.
 *
 * Revision 1.32  2002/10/07 11:52:27  gerben
 * Several fixes related to possible invalid content.
 *
 * Revision 1.30  2002/10/07 08:35:57  gerben
 * Aborting a task is now understood and reported correctly.
 *
 * Revision 1.29  2002/10/06 10:19:20  gerben
 * Wrote workaround for indentation problem (indentation stays around even when view is cleared).
 *
 * Revision 1.28  2002/10/05 19:43:08  gerben
 * Give feedback when install/etc end succesfully.
 *
 * Revision 1.27  2002/10/03 18:15:21  gerben
 * A typo (bz instead of bz2) prevented the unarchive of anything.
 *
 * Revision 1.26  2002/10/03 13:38:03  gerben
 * Removed addChildWindow as child passes close to parent. Window can only close when there are no running tasks.
 *
 * Revision 1.25  2002/10/02 22:38:30  gerben
 * Added CVS Log in C Comment at the end of every file
 *
 */