//  Copyright (c) 2002-2004, Gerben Wierda, R&A
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
//
//  Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
//  Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
//
//  Neither the name of R&A nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#import <Cocoa/Cocoa.h>

@class II2TaskSet;

@protocol II2TaskSetObserverProtocol
// Sent to the observer whenever the task set changes
- (void)ii2TaskSetDidRefresh:(II2TaskSet *)sender;
@end

@protocol II2TaskDependencyProtocol <NSObject>
- (NSString *)description;
- (void)wait;
- (void)endWait;
- (void)abortWait;
- (void)addDependentTask:(id <II2TaskDependencyProtocol>)task;
- (NSArray *)dependentTasks;
- (void)addDependentTasks:(NSArray *)tasks;
@end

@protocol II2TaskStatusProviderProtocol;

@protocol II2TaskStatusReporterProtocol <NSObject>
- (void)updateReportWithDataFromTask:(id <II2TaskStatusProviderProtocol>)task download:(BOOL)download;
- (void)setReportMaxDownloadsSize:(int)size;
- (void)setReportMaxActionsSize:(int)size;
@end

@protocol II2TaskStatusProviderProtocol <NSObject>
- (void)setReportTarget:(id <II2TaskStatusReporterProtocol>)reporter;
- (void)setReportIncrementTreshold:(int)treshold;
- (NSString *)typeOfTask;
- (int)sizeOfTaskWhenCompleted;
- (int)currentSizeOfTask;
- (int)sizeOfTaskSinceLastReported;
@end

@protocol II2TaskSetMemberProtocol <II2TaskDependencyProtocol>
// Tasks in this set will have to conform to this protocol
- (NSAttributedString *)attributedDisplayName;
- (id)progress;
- (void)abort;
- (void)launch;
- (void)setTaskSet:(II2TaskSet *)set;
@end

@protocol II2TaskSetHandlerProtocol
// When comparing the object argument of the target/action for equality, the handler should know how to compare as it did put the object there in the first place
- (BOOL)ii2TaskObject: (id)object1 isEqualTo: (id)object2;
// Called by e.g. observer to trigger aborting a task in the set:
- (void)abortTaskAtIndex:(int)index;
@end

@protocol II2TaskPropertiesProtocol
- (void)setProperties: (NSDictionary *)myProperties;
- (NSDictionary *)properties;
- (id)propertyForKey:(id <NSCopying>)key;
- (void)setProperty:(id)property forKey:(id <NSCopying>)key;
@end

@interface II2TaskSet : NSObject {
    id <II2TaskSetObserverProtocol>	__observer;
    id <II2TaskSetHandlerProtocol>	__handler;

    NSMutableArray			*__tasks;
}

- (void)addII2Task:(id <II2TaskSetMemberProtocol>)task;

- (void)ii2TaskDidEnd:(id <II2TaskSetMemberProtocol>)task withSuccess:(BOOL)success;

// This observer is sent ii2TaskSetDidRefresh:self when the table changes. A window controller probably
- (void)setObserver:(id <II2TaskSetObserverProtocol>)observer;

// This observer is sent ii2TaskEnded:fromSet:self when a task is removed from the set, a document probably
- (void)setHandler:(id <II2TaskSetHandlerProtocol>)handler;

// We can report the number of running tasks
- (unsigned) count;

// We can report the task at a row index:
- (id <II2TaskSetMemberProtocol>)taskAtIndex:(int)row;

// We can act as table source for NSTableView:
- (int)numberOfRowsInTableView:(NSTableView *)aTableView;
- (id)tableView:(NSTableView *)aTableView objectValueForTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex;

- (void)refreshDisplay:(id <II2TaskSetMemberProtocol>)task;

@end

/*
 * $Log: II2TaskSet.h,v $
 * Revision 1.3  2004/04/25 13:13:21  gctwnl
 * Forced commit of all .h and .m files of version 2.61.0
 *
 * Revision 1.2  2004/03/19 23:35:11  gctwnl
 * Version 2.55.0. Has a renewed interface to support simple operations.
 * Supports progress bar control from the i-Package.
 *
 * Revision 1.1.1.1  2004/02/01 17:18:27  gctwnl
 * Initial import.
 *
 * Revision 3.3  2004/01/14 12:56:46  gerben
 * Version 2.39.0
 *
 * Revision 3.2  2002/12/10 14:24:54  gerben
 * Set sticky tag 3.2
 *
 * Revision 3.0  2002/12/10 14:04:24  gerben
 * Upping all files with a sticky tag of 1.99
 *
 * Revision 1.99  2002/12/10 13:52:53  gerben
 * Set all revision numbers to 1.99
 *
 * Revision 1.12  2002/10/22 14:14:20  gerben
 * Differ between different user actions.
 *
 * Revision 1.11  2002/10/16 21:28:22  gerben
 * A complete new rewrite of taskss, processes and stuff.
 *
 * Revision 1.10  2002/10/14 12:34:19  gerben
 * Removed dead code
 *
 * Revision 1.9  2002/10/12 10:54:38  gerben
 * A lot of changes, mostly when producing output/feedback.
 *
 * Revision 1.8  2002/10/07 20:28:06  gerben
 * Removed circular CVS Log entry in C Comments at the end of files
 *
 * Revision 1.7  2002/10/03 18:15:22  gerben
 * A typo (bz instead of bz2) prevented the unarchive of anything.
 *
 * Revision 1.6  2002/10/02 22:38:31  gerben
 * Added CVS Log in C Comment at the end of every file
 *
 */