/*
 * Copyright (c) 2002-2004, Gerben Wierda, R&A All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * Neither the name of R&A nor the names of its contributors may be used to
 * endorse or promote products derived from this software without specific prior
 * written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * This is gluprunner, Gerben's Leader-setting, User-setting, Process-
 *  reporting runner program
 * (c) Gerben Wierda, 2002
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/param.h>
#include <sys/wait.h>
extern int errno;

main( int argc, char *argv[], char *envp[]) {
    extern char **environ; /* works, where passing on envp does not work (envp 0??) */
    const char *progname = argv[0];
    char envsizebuf[32];
    int i;
    int ch;
    int quiet = 0;
    const char *pfifoname = NULL; /* pid fifo name */
    const char *ififoname = NULL; /* stdin fifo name */
    const char *ofifoname = NULL; /* stdout fifo name */
    const char *efifoname = NULL; /* stderr fifo name */
    const char *Efifoname = NULL; /* stderr fifo name */
    int ififo=-1, ofifo=-1, efifo=-1, pfifo=-1, Efifo=-1;
    pid_t pid = getpid();

    setuid( geteuid());
    setpgrp( 0, setsid());

    /*
     * getopt handling:
     *  -d <dir>
     *	    Change working directory to <dir> before execution
     *	-i <file>
     *	    Use <file> as the fifo to connect stdin to
     *	-o <file>
     *	    Use <file> as the fifo to connect stdout to
     *  -e <file>
     *	    Use <file> as the fifo to connect stderr to
     *	-E <file>
     *	    Read environment settings from fifo <file>
     *	-q
     *	    Quiet. Do not print PID
     *  -p <file>
     *	    Print pid (raw) to fifo <file>
     */

    while ((ch = getopt( argc, argv, "qd:o:e:p:i:E:")) != -1) {
	switch (ch) {
	case 'i':
	    if ((ififo = open( (ififoname = optarg), O_RDONLY, 0)) < 0) {
		fprintf(stderr, "%s: open %s: %s\n", progname, optarg, strerror( errno));
		exit(1);
	    }
	    break;
	case 'o':
	    if ((ofifo = open( (ofifoname = optarg), O_WRONLY|O_APPEND, 0)) < 0) {
		fprintf(stderr, "%s: open %s: %s\n", progname, optarg, strerror( errno));
		exit(1);
	    }
	    break;
	case 'e':
	    if ((efifo = open( (efifoname = optarg), O_WRONLY|O_APPEND, 0)) < 0) {
		fprintf(stderr, "%s: open %s: %s\n", progname, optarg, strerror( errno));
		exit(1);
	    }
	    break;
	case 'p':
	    if ((pfifo = open( (pfifoname = optarg), O_WRONLY|O_APPEND, 0)) < 0) {
		fprintf(stderr, "%s: open %s: %s\n", progname, optarg, strerror( errno));
		exit(1);
	    }
	    write( pfifo, &pid, sizeof( pid_t));
	    close( pfifo);
	    break;
	case 'E':
	    if ((Efifo = open( (Efifoname = optarg), O_RDONLY, 0)) < 0) {
		fprintf(stderr, "%s: open %s: %s\n", progname, optarg, strerror( errno));
		exit(1);
	    }
	    i=0;
	    while (i<sizeof(envsizebuf)) {
		if (read( Efifo, envsizebuf+i, 1) < 0) {
		    fprintf(stderr, "%s: read %s: %s\n", progname, optarg, strerror( errno));
		    exit( 1);
		}
		if (envsizebuf[i] == '\n') {
		    break;
		}
		i++;
	    }
	    if (i<sizeof(envsizebuf)) {
		unsigned int envsize;
		char *envblock, *startpt, *endpt;
		if (sscanf( envsizebuf, "%u", &envsize) != 1) {
		    fprintf(stderr, "%s: sscanf %s: %s\n", progname, optarg, envsizebuf);
		    exit( 1);
		}
		envblock = calloc( envsize, 1);
		if (read( Efifo, envblock, envsize) < envsize) {
		    fprintf(stderr, "%s: read block too small %s: %u\n", progname, optarg, envsize);
		    exit( 1);
		}
		startpt = envblock;
		while (startpt < (envblock + envsize)) {
		    endpt = strchr( startpt, '\n');
		    *endpt = 0;
		    /* fprintf( stderr, "ENV: %s\n", startpt); fflush( stderr); */
		    if (putenv( startpt)) {
			fprintf(stderr, "%s: putenv failure %s: %s\n", progname, startpt, strerror( errno));
		    }
		    startpt = endpt+1;
		}
		free( envblock);
	    }
	    close( Efifo);
	    unlink( Efifoname);
	    break;
	case 'd':
	    if (chdir( optarg) != 0) {
		fprintf( stderr, "%s: chdir %s: %s\n", progname, optarg, strerror( errno));
		fflush( stderr);
		exit(1);
	    }
	    break;
	case 'q':
	    quiet = 1;
	    break;
	default:
	    fprintf(stderr, "%s: illegal option\n", progname);
	    exit( 1);
	}
    }
    argc -= (optind-1);
    argv += (optind-1);

    if (ififo != -1) {
	if (dup2( ififo, fileno(stdin)) == -1) {
	    fprintf(stderr, "%s: Cannot dup2 %u: %s\n", progname, fileno( stdin), strerror( errno));
	    exit(1);
	}
    }
    if (ofifo != -1) {
	if (dup2( ofifo, fileno(stdout)) == -1) {
	    fprintf(stderr, "%s: Cannot dup2 %u: %s\n", progname, fileno( stdout), strerror( errno));
	    exit(1);
	}
    }
    if (efifo != -1) {
	if (dup2( efifo, fileno(stderr)) == -1) {
	    fprintf(stderr, "%s: Cannot dup2 %u: %s\n", progname, fileno( stderr), strerror( errno));
	    exit(1);
	}
    }

    if ((pfifo == -1) && !quiet) {
	fprintf( stdout, "%s: PID = %u\n", progname, pid); fflush( stdout);
    }

    execve( argv[1], &argv[1], environ);
}
